"""SCVx* algorithm"""

import copy
import numpy as np
import time


class SCPSolution:
    """Solution structure for SCvx* algorithm

    The solution structure is to the non-convex OCP of the form:

    min_{x,u,v,y} J(x,u,v,y)
    s.t. 
        x_{k+1} = x_k + f(x_k,u_k,v_k,y_k)
        ||u_k|| <= v_k
        g(x,u,v,y) = 0
        h(x,u,v,y) <= 0
    
    Attributes:
        x (np.array): optimal state trajectory
        u (np.array): optimal control trajectory
        v (np.array): optimal control magnitude terms
        y (np.array): optimal general variables
        sols (list): list of solutions
        summary_dict (dict): summary dictionary
    """
    def __init__(self, xopt, uopt, vopt, yopt, sols, summary_dict, variables_each_iterations=None):
        self.x = xopt
        self.u = uopt
        self.v = vopt
        self.y = yopt
        self.sols = sols
        self.summary_dict = summary_dict
        self.variables_each_iterations = variables_each_iterations
        return


class SCvxStar:
    """SCvx* algorithm for optimal control problems

    Hyperparameters are defined according to `Oguri, 2023` (doi: 10.1109/CDC49753.2023.10383462).

    Args:
        problem (ContinuousControlSCOCP or ImpulsiveControlSCOCP): `SCOCP` instance
        tol_opt (float): optimality tolerance
        tol_feas (float): feasibility tolerance
        rho0 (float): step acceptance criterion parameter
        rho1 (float): trust-region radius contraction threshold
        rho2 (float): trust-region radius expansion threshold
        alpha1 (float): trust-region radius contraction factor s.t. r_k+1 = max(r_k/alpha1, r_bounds[0])
        alpha2 (float): trust-region radius expansion factor s.t. r_k+1 = min(r_k*alpha2, r_bounds[1])
        beta (float): weight update factor
        gamma (float): update factor forLagrange multiplier update criterion delta
        r_bounds (list): trust region bounds
        weight_max (float): maximum weight
    """
    def __init__(
        self,
        problem,
        tol_opt = 1e-6,
        tol_feas = 1e-6,
        rho0 = 0.0,
        rho1 = 0.25,
        rho2 = 0.7,
        alpha1 = 2.0,
        alpha2 = 3.0,
        beta = 2.0,
        gamma = 0.9,
        r_bounds = [1e-8, 10.0],
        steps_minimum_trust_region = 10,
        weight_max = 1e16,
    ):
        # assertions on hyperparameters
        assert 0.0 <= rho0 < 1.0, "rho0 must be in [0.0, 1.0)"
        assert alpha1 > 1.0, "alpha1 must be greater than 1.0"
        assert alpha2 > 1.0, "alpha2 must be greater than 1.0"
        assert rho1 < rho2, "rho1 must be less than rho2"
        assert beta > 1.0, "beta must be greater than 1.0"
        assert 0.0 < gamma < 1.0, "gamma must be in (0.0, 1.0)"
        assert r_bounds[0] > 0.0, "r_bounds[0] must be greater than 0.0"
        assert r_bounds[1] > r_bounds[0], "r_bounds[1] must be greater than r_bounds[0]"
        
        self.problem = problem
        self.tol_opt = tol_opt
        self.tol_feas = tol_feas
        self.rho0 = rho0
        self.rho1 = rho1
        self.rho2 = rho2
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.beta = beta
        self.gamma = gamma
        self.r_bounds = r_bounds
        self.steps_minimum_trust_region = steps_minimum_trust_region
        self.weight_max = weight_max
        return
    

    def evaluate_penalty(self, gdyn, g, h):
        """Evaluate penalty function according to Augmented Lagrangian formulation
        
        Args:
            gdyn (np.array): (N-1)-by-nx array of nonlinear dynamics constraints violations
            g (np.array): ng-by-1 array of nonlinear equality constraints violations
            h (np.array): nh-by-1 array of nonlinear inequality constraints violations
        """
        assert gdyn.shape == (self.problem.N-1, self.problem.integrator.nx),\
            f"gdyn.shape = {gdyn.shape} != (self.problem.N-1, self.problem.integrator.nx) = {(self.problem.N-1, self.problem.integrator.nx)}"
        Nseg,_ = self.problem.lmb_dynamics.shape
        penalty = 0.0
        for i in range(Nseg):
            penalty += self.problem.lmb_dynamics[i,:] @ gdyn[i,:] + self.problem.weight/2 * gdyn[i,:] @ gdyn[i,:]
        if self.problem.ng > 0:
            assert g.shape == (self.problem.ng,)
            penalty += self.problem.lmb_eq @ g + self.problem.weight/2 * (g @ g)
        if self.problem.nh > 0:
            assert h.shape == (self.problem.nh,)
            h_pos = np.clip(h, 0, None)      # penalize only when inequality constraint is violated
            penalty += self.problem.lmb_ineq @ h_pos + self.problem.weight/2 * (h_pos @ h_pos)
        return penalty
    
    
    def solve(
        self,
        xbar,
        ubar,
        vbar = None,
        ybar = None,
        maxiter: int = 10,
        verbose: bool = True,
        feasability_norm = np.inf,
        debug = False,
        exit_condition_rho = False,
        save_all_iterations = False,
    ):
        """Solve optimal control problem via SCvx* algorithm
        
        Args:
            xbar (np.array): N-by-nx array of reference states
            ubar (np.array): N-by-nu array of reference controls
            vbar (np.array): N-by-1 array of reference constraints
            maxiter (int): maximum number of iterations
            verbose (bool): whether to print verbose output
            feasability_norm (str): norm to use for feasibility evaluation
            debug (bool): whether to print debug output
            exit_condition_rho (bool): whether to exit if rho < rho0
            save_all_iterations (bool): whether to save all iterations

        Returns:
            SCPSolution: solution object
        """
        tstart = time.time()
        header = f"|  Iter  |     J0      |   Delta J   |   Delta L   |    chi     |     rho     |     r      |   weight   | step acpt. |"
        print_frequency = 10
        delta = 1e16
        status_AL = "NotConverged"
        chi = 1e15
        rho = self.rho0  # initialize rho to rho0
        n_min_trust_region = 0
        sols = []

        # initialize vbar if not provided
        if vbar is None:
            assert self.problem.augment_Gamma == False, f"If augment_Gamma is True, vbar must be provided"
            # vbar = np.sum(ubar, axis=1).reshape(-1,1)
        if ybar is None and self.problem.ny > 0:
            ybar = np.zeros((self.problem.ny,))

        # initial constraint violation evaluation
        gdyn_nl_bar, _ = self.problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar)
        g_nl_bar, h_nl_bar = self.problem.evaluate_nonlinear_constraints(xbar, ubar, vbar, ybar)
        assert g_nl_bar.shape == (self.problem.ng,),\
            f"Shape of equality constraint violations by self.problem.evaluate_nonlinear_constraints does not match (self.problem.ng,)"
        assert h_nl_bar.shape == (self.problem.nh,),\
            f"Shape of inequality constraint violations by self.problem.evaluate_nonlinear_constraints does not match (self.problem.nh,)"

        # initialize summary dictionary
        scp_summary_dict = {
            "num_iter": 0,
            "status": "NotConverged",
            "J0": [],
            "chi": [],
            "DeltaJ": [],
            "DeltaL": [],
            "accept": [],
            "weight": self.problem.weight,
            "trust_region_radius": self.problem.trust_region_radius,
            "rho": self.rho0,
            "t_cvx": [],
            "t_scp": [],
        }
        variables_each_iterations = []

        for k in range(maxiter):
            tstart_scp = time.time()

            # build linear model
            self.problem.build_linear_model(xbar, ubar, vbar)
            tstart_cvx = time.time()
            xopt, uopt, vopt, yopt, xi_dyn_opt, xi_opt, zeta_opt = self.problem.solve_convex_problem(xbar, ubar, vbar, ybar)
            scp_summary_dict["t_cvx"].append(time.time() - tstart_cvx)
            if self.problem.cp_status not in ["optimal", "optimal_inaccurate"]:
                status_AL = "CPFailed"
                if verbose:
                    print(f"    Convex problem did not converge to optimality (status = {self.problem.cp_status})!")
                break
            
            # evaluate nonlinear dynamics
            gdyn_nl_opt, sols = self.problem.evaluate_nonlinear_dynamics(xopt, uopt, vopt)

            # evaluate nonlinear constraints
            g_nl_opt, h_nl_opt = self.problem.evaluate_nonlinear_constraints(xopt, uopt, vopt, yopt)
            chi = np.linalg.norm(np.concatenate((gdyn_nl_opt.flatten(order='C'), g_nl_opt, h_nl_opt)), feasability_norm)

            # evaluate penalized objective
            J0     = self.problem.evaluate_objective(xopt, uopt, vopt, yopt)
            J0_bar = self.problem.evaluate_objective(xbar, ubar, vbar, ybar)
            J_bar  = J0_bar + self.evaluate_penalty(gdyn_nl_bar, g_nl_bar, h_nl_bar)
            J_opt  = J0     + self.evaluate_penalty(gdyn_nl_opt, g_nl_opt, h_nl_opt)
            L_opt  = J0     + self.evaluate_penalty(xi_dyn_opt, xi_opt, zeta_opt)

            # evaluate step acceptance criterion parameter
            DeltaJ = J_bar - J_opt
            DeltaL = J_bar - L_opt
            rho = DeltaJ / DeltaL

            if debug:
                print(f"\n  SCvxStar debug mode output at iteration {k+1}:")
                print(f"    J0            = {J0:1.4e}")
                print(f"    P(g(z), h(z)) = {self.evaluate_penalty(gdyn_nl_opt, g_nl_opt, h_nl_opt):1.4e}")
                print(f"    P(xi, zeta)   = {self.evaluate_penalty(xi_dyn_opt, xi_opt, zeta_opt):1.4e}")
                print(f"        ||gdyn_nl_opt|| = {np.linalg.norm(gdyn_nl_opt, np.inf):1.4e}, ||xi_dyn_opt||  = {np.linalg.norm(xi_dyn_opt, np.inf):1.4e}")
                if self.problem.ng > 0:
                    print(f"        ||g_nl_opt||    = {np.linalg.norm(g_nl_opt, np.inf):1.4e}, ||xi_opt||      = {np.linalg.norm(xi_opt, np.inf):1.4e}")
                if self.problem.nh > 0:
                    print(f"        ||h_nl_opt||    = {np.linalg.norm(h_nl_opt, np.inf):1.4e}, ||zeta_opt||      = {np.linalg.norm(zeta_opt, np.inf):1.4e}")
                print(f"    J0 = {J0:1.4e}, J0_bar = {J0_bar:1.4e}")
                print(f"    J_opt = J0 + P(g(z), h(z)) = {J0:1.4e} + {self.evaluate_penalty(gdyn_nl_opt, g_nl_opt, h_nl_opt):1.4e} = {J_opt:1.4e}")
                print(f"    L_opt = J0 + P(xi, zeta)   = {J0:1.4e} + {self.evaluate_penalty(xi_dyn_opt, xi_opt, zeta_opt):1.4e} = {L_opt:1.4e}")
                print(f"    Delta J = {DeltaJ:1.4e}, Delta L = {DeltaL:1.4e}, rho = {rho:1.4e}")
                # print(f"        xi_opt              = {xi_opt}")
                # print(f"        self.problem.lmb_eq = {self.problem.lmb_eq}")
                print(f"    chi = {chi:1.4e}")
                print("\n")

            if rho >= self.rho0:
                step_acpt_msg = "yes"
            else:
                step_acpt_msg = "no "
            if verbose:
                if np.mod(k, print_frequency) == 0:
                    print(f"\n{header}")
                print(f"   {k+1:3d}   | {J0: 1.4e} | {DeltaJ: 1.4e} | {DeltaL: 1.4e} | {chi:1.4e} | {rho: 1.4e} | {self.problem.trust_region_radius:1.4e} | {self.problem.weight:1.4e} |    {step_acpt_msg}     |")

            # update storage
            scp_summary_dict["J0"].append(J0)
            scp_summary_dict["chi"].append(chi)
            scp_summary_dict["DeltaJ"].append(DeltaJ)
            scp_summary_dict["DeltaL"].append(DeltaL)
            scp_summary_dict["accept"].append(int(rho >= self.rho0))
            if save_all_iterations:
                variables_each_iterations.append({
                    "x": copy.deepcopy(xopt),
                    "u": copy.deepcopy(uopt),
                    "v": copy.deepcopy(vopt),
                    "y": copy.deepcopy(yopt),
                })

            if (chi <= self.tol_feas) and (abs(DeltaJ) <= self.tol_opt) and (exit_condition_rho is False or rho >= self.rho0):
                status_AL = "Optimal"
                break
                
            if rho >= self.rho0:
                xbar[:,:] = xopt[:,:]
                ubar[:,:] = uopt[:,:]
                if vbar is not None:
                    vbar[:,:] = vopt[:,:]
                if self.problem.ny > 0:
                    ybar[:] = yopt[:]
                gdyn_nl_bar[:,:] = gdyn_nl_opt[:,:]
                if self.problem.ng > 0:
                    g_nl_bar[:] = g_nl_opt[:]
                if self.problem.nh > 0:
                    h_nl_bar[:] = h_nl_opt[:]

                if abs(DeltaJ) < np.abs(delta):
                    # update multipliers
                    self.problem.lmb_dynamics = self.problem.lmb_dynamics + self.problem.weight * gdyn_nl_opt
                    if self.problem.ng > 0:
                        self.problem.lmb_eq   = self.problem.lmb_eq + self.problem.weight * g_nl_opt
                    if self.problem.nh > 0:
                        self.problem.lmb_ineq = np.clip(self.problem.lmb_ineq + self.problem.weight * h_nl_opt, 0, None)

                    # update weight
                    self.problem.weight = min(self.beta * self.problem.weight, self.weight_max)
                    
                    # multiplier & weight update
                    if delta > 1e15:
                        delta = abs(DeltaJ)
                    else:
                        delta *= self.gamma

            # update trust-region
            if rho < self.rho1:
                self.problem.trust_region_radius = max(self.problem.trust_region_radius/self.alpha1, self.r_bounds[0])
            elif rho >= self.rho2:
                self.problem.trust_region_radius = min(self.problem.trust_region_radius*self.alpha2, self.r_bounds[1])

            # update steps at minimum trust region
            if self.problem.trust_region_radius == self.r_bounds[0]:
                n_min_trust_region += 1
            else:
                n_min_trust_region = 0
            if n_min_trust_region >= self.steps_minimum_trust_region:
                status_AL = "SlowProgress"
                if verbose:
                    print(f"    Exceeded {n_min_trust_region} steps at minimum trust region! Stopping SCvx*due to slow progress...\n")
                break
            scp_summary_dict["t_scp"].append(time.time() - tstart_scp)

        if (k == maxiter - 1) and (status_AL not in ["Optimal", "CPFailed", "SlowProgress"]):
            if chi <= self.tol_feas:
                status_AL = "Feasible"
            else:
                status_AL = "MaxIter"

        # print summary
        t_algorithm = time.time() - tstart
        if verbose:
            print("\n")
            print(f"    SCvx* algorithm summary:")
            print(f"        Status                          : {status_AL}")
            print(f"        Objective value                 : {scp_summary_dict['J0'][-1]:1.8e}")
            print(f"        Penalized objective improvement : {scp_summary_dict['DeltaJ'][-1]:1.8e} (tol: {self.tol_opt:1.4e})")
            print(f"        Constraint violation            : {scp_summary_dict['chi'][-1]:1.8e} (tol: {self.tol_feas:1.4e})")
            print(f"        Total iterations                : {k+1}")
            print(f"        SCvx* algorithm time            : {t_algorithm:1.4f} seconds")
            print("\n")

        # update summary dictionary
        scp_summary_dict["num_iter"] = k + 1
        scp_summary_dict["status"] = status_AL
        scp_summary_dict["status_CP"] = self.problem.cp_status
        scp_summary_dict["weight"] = self.problem.weight
        scp_summary_dict["trust_region_radius"] = self.problem.trust_region_radius
        scp_summary_dict["rho"] = rho
        scp_summary_dict["t_algorithm"] = t_algorithm
        return SCPSolution(xopt, uopt, vopt, yopt, sols, scp_summary_dict, variables_each_iterations)
    
    def plot_DeltaJ(self, axis, summary_dict: dict, s = 5):
        """Plot iterations of DeltaJ"""
        iters = np.arange(len(summary_dict["DeltaJ"]))
        axis.plot(iters, np.abs(summary_dict["DeltaJ"]), c='k', lw=0.5)
        axis.scatter(iters, np.abs(summary_dict["DeltaJ"]), marker="o", s=s, color=['g' if a == 1 else 'r' for a in summary_dict["accept"]], zorder=2)
        axis.set(yscale='log', xlabel='Iter.', ylabel='|DeltaJ|')
        return
    
    def plot_chi(self, axis, summary_dict: dict, s = 5):
        """Plot iterations of chi"""
        iters = np.arange(len(summary_dict["chi"]))
        axis.plot(iters, summary_dict["chi"], c='k', lw=0.5)
        axis.scatter(iters, summary_dict["chi"], marker="o", s=s, color=['g' if a == 1 else 'r' for a in summary_dict["accept"]], zorder=2)
        axis.set(yscale='log', xlabel='Iter.', ylabel='chi')
        return
    
    def plot_J0(self, axis, summary_dict: dict, s = 5):
        """Plot iterations of J0"""
        iters = np.arange(len(summary_dict["J0"]))
        axis.plot(iters, summary_dict["J0"], c='k', lw=0.5)
        axis.scatter(iters, summary_dict["J0"], marker="o", s=s, color=['g' if a == 1 else 'r' for a in summary_dict["accept"]], zorder=2)
        axis.set(xlabel='Iter.', ylabel='J0')
        return



"""Equations of motion library"""

from ._twobody_scipy import (
    gravity_gradient_twobody,
    rhs_twobody,
    rhs_twobody_stm,
    control_rhs_twobody_logmass_freetf,
    control_rhs_twobody_logmass_freetf_stm,
    control_rhs_twobody_mass_freetf,
    control_rhs_twobody_mass_freetf_stm,
)

from ._cr3bp_scipy import (
    gravity_gradient_cr3bp,
    rhs_cr3bp,
    rhs_cr3bp_stm,
    control_rhs_cr3bp,
    control_rhs_cr3bp_stm,
    control_rhs_cr3bp_logmass,
    control_rhs_cr3bp_logmass_stm,
    control_rhs_cr3bp_freetf,
    control_rhs_cr3bp_freetf_stm,
    control_rhs_cr3bp_logmass_freetf,
    control_rhs_cr3bp_logmass_freetf_stm,
)


"""Two-body dynamics"""

from numba import njit, float64
import numpy as np


@njit(float64[:,:](float64[:], float64), cache=True)
def gravity_gradient_twobody(rvec, mu):
    """Compute gravity gradient matrix for R2BP in the rotating frame.

    Args:
        mu (float): R2BP parameter
        rvec (np.array): position vector of spacecraft

    Returns:
        np.array: 3-by-3 gravity gradient matrix
    """
    # unpack rvec
    x,y,z = rvec
    rnorm = np.sqrt(x**2 + y**2 + z**2)
    G = mu/rnorm**5 * (3*np.outer(rvec, rvec) - np.eye(3)*rnorm**2)
    return G


@njit(float64[:](float64, float64[:], float64), cache=True)
def rhs_twobody(t, state, mu):
    """Equation of motion in R2BP, formulated for scipy.integrate.solve=ivp(), compatible with njit

    Args:
        t (float): time
        state (np.array): 1D array of Cartesian state, length 6
        mu (float): R2BP parameter

    Returns:
        (np.array): 1D array of derivative of Cartesian state
    """
    deriv = np.zeros((6,))
    deriv[0:3] = state[3:6]
    deriv[3:6] = -mu * state[0:3] / np.linalg.norm(state[0:3])**3
    return deriv


@njit(float64[:](float64, float64[:], float64), cache=True)
def rhs_twobody_stm(t, state, mu):
    """Equation of motion in R2BP in the rotating frame with STM.
    This function is written for `scipy.integrate.solve=ivp()` and is compatible with njit.
    """
    # state derivative
    deriv = np.zeros(12)
    deriv[0:6] = rhs_twobody(t, state[0:6], mu)
    
    # derivative of STM
    A = np.zeros((6,6))
    A[0:3,3:6] = np.eye(3)
    A[3:6,0:3] = gravity_gradient_twobody(state[0:3], mu)
    # Create a contiguous copy of the STM part before reshaping
    stm = np.copy(state[6:])
    deriv[6:] = np.dot(A, stm.reshape(6,6)).reshape(36,)
    return deriv


def control_rhs_twobody_logmass_freetf(tau, state, parameters, u):
    """Equation of motion in R2BP with continuous control in the rotating frame
    state = [x,y,z,vx,vy,vz,log(mass),t]
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    """
    # unpack parameters
    mu, cex = parameters
    # derivative of state
    deriv = np.zeros(8,)
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv[0:6] = u[3] * (rhs_twobody(state[6], state[0:6], mu) + B @ u[0:3])
    deriv[6]   = u[3] * u[4] * (-1/cex)     # control on log(mass), multiplied by time-dilation factor
    deriv[7]   = u[3]                       # control on t(tau)
    return deriv


def control_rhs_twobody_logmass_freetf_stm(tau, state, parameters, u):
    """Equation of motion in R2BP with continuous control in the rotating frame with STM
    state = [x,y,z,vx,vy,vz,log(mass),t] + STM.flatten()
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    """
    # derivative of state
    mu, cex = parameters
    deriv = np.zeros(112)    # 8 + 8*8 + 8*5
    deriv[0:8] = control_rhs_twobody_logmass_freetf(tau, state[0:8], parameters, u)
    
    # derivative of STM
    Phi_A = state[8:72].reshape(8,8)
    A = np.zeros((8,8))
    A[0:3,3:6] = np.eye(3)
    A[3:6,0:3] = gravity_gradient_twobody(state[0:3], mu)
    deriv[8:72] = np.dot(u[3] * A, Phi_A).reshape(64,)

    # derivative of control sensitivity
    B = np.zeros((8,5))
    B[3:6,0:3] = u[3] * np.eye(3)
    B[0:8,3] = control_rhs_twobody_logmass_freetf(tau, state[0:8], parameters, u)/u[3]
    B[6,4] = -u[3]/cex
    Phi_B = state[72:112].reshape(8,5)
    deriv[72:112] = (np.dot(u[3] * A, Phi_B) + B).reshape(40,)
    return deriv


def control_rhs_twobody_mass_freetf(tau, state, parameters, u):
    """Equation of motion in R2BP with continuous control in the rotating frame
    
    - state = [x,y,z,vx,vy,vz,mass,t]
    - u = [ux, uy, uz, s, Gamma] where s is the dilation factor, Gamma is the control throttle magnitude (at convergence)
    - parameters = [mu, c1, c2]
    """
    # unpack parameters
    mu, c1, c2 = parameters
    # derivative of state
    deriv = np.zeros(8,)
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv[0:6] = u[3] * (rhs_twobody(state[6], state[0:6], mu) + B @ u[0:3] * (c1/state[6]))
    deriv[6]   = u[3] * u[4] * (-c2)     # control on log(mass), multiplied by time-dilation factor
    deriv[7]   = u[3]                       # control on t(tau)
    return deriv


def control_rhs_twobody_mass_freetf_stm(tau, state, parameters, u):
    """Equation of motion in R2BP with continuous control in the rotating frame with STM

    - state = [x,y,z,vx,vy,vz,mass,t] + STM.flatten()
    - u = [ux, uy, uz, s, Gamma] where s is the dilation factor, Gamma is the control throttle magnitude (at convergence)
    - parameters = [mu, c1, c2]
    """
    # derivative of state
    mu, c1, c2 = parameters
    deriv = np.zeros(112)    # 8 + 8*8 + 8*5
    deriv[0:8] = control_rhs_twobody_mass_freetf(tau, state[0:8], parameters, u)
    
    # derivative of STM
    Phi_A = state[8:72].reshape(8,8)
    A = np.zeros((8,8))
    A[0:3,3:6] = np.eye(3)
    A[3:6,0:3] = gravity_gradient_twobody(state[0:3], mu)
    A[3:6,6]   = -(c1/state[6]**2) * np.array(u[0:3])
    deriv[8:72] = np.dot(u[3] * A, Phi_A).reshape(64,)

    # derivative of control sensitivity
    B = np.zeros((8,5))
    B[3:6,0:3] = u[3] * c1/state[6] * np.eye(3)
    B[0:8,3] = control_rhs_twobody_mass_freetf(tau, state[0:8], parameters, u)/u[3]
    B[6,4] = -u[3] * c2
    Phi_B = state[72:112].reshape(8,5)
    deriv[72:112] = (np.dot(u[3] * A, Phi_B) + B).reshape(40,)
    return deriv


"""Free-time problems with moving targets"""

from collections.abc import Callable
import cvxpy as cp
import numpy as np

from ._scocp_continuous import ContinuousControlSCOCP
from .._misc import get_augmented_lagrangian_penalty, MovingTarget


class FreeTimeContinuousMovingTargetRdvLogMass(ContinuousControlSCOCP):
    """Free-time continuous moving target rendezvous problem class with log-mass dynamics
    
    Note the ordering expected for the state and the control vectors: 

    state = [x,y,z,vx,vy,vz,log(mass),t]
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    
    The target state is given by a `MovingTarget` object, which result in 6 non-convex equality constraints.
    """
    def __init__(self, x0, target: MovingTarget, Tmax, tf_bounds, s_bounds, N, *args, **kwargs):
        super().__init__(ng=6, nh = N - 1, *args, **kwargs)
        assert len(x0) == 7
        assert abs(self.times[0]  - 0.0) < 1e-14, f"self.times[0] must be 0.0, but given {self.times[0]}"
        assert abs(self.times[-1] - 1.0) < 1e-14, f"self.times[-1] must be 1.0, but given {self.times[-1]}"
        assert s_bounds[0] > 0.0, f"s_bounds[0] must be greater than 0.0, but given {s_bounds[0]}"
        assert s_bounds[0] < s_bounds[1], f"s_bounds[0] must be less than s_bounds[1], but given {s_bounds[0]} and {s_bounds[1]}"
        self.x0 = x0
        self.target = target
        self.Tmax = Tmax
        self.tf_bounds = tf_bounds
        self.s_bounds = s_bounds
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        return -xs[-1,6]
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, gs, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        
        xs      = cp.Variable((N, nx), name='state')
        us      = cp.Variable((Nseg, nu), name='control')
        vs      = cp.Variable((Nseg, 1), name='Gamma')
        xis_dyn = cp.Variable((Nseg,nx), name='xi_dyn')         # slack for dynamics
        xis     = cp.Variable((self.ng,), name='xi')         # slack for target state
        zetas   = cp.Variable((Nseg,), name='zeta')     # slack for non-convex inequality
        
        penalty = get_augmented_lagrangian_penalty(
            self.weight,
            xis_dyn,
            self.lmb_dynamics,
            xi=xis,
            lmb_eq=self.lmb_eq,
            zeta=zetas,
            lmb_ineq=self.lmb_ineq
        )
        objective_func = -xs[-1,6] + penalty
        constraints_objsoc = [cp.SOC(vs[i,0], us[i,0:3]) for i in range(N-1)]
        
        constraints_dyn = [
            xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,0:4] @ us[i,:] + self.Phi_B[i,:,4] * vs[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
            for i in range(Nseg)
        ]

        constraints_trustregion = [
            xs[i,:] - xbar[i,:] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,:] - xbar[i,:] >= -self.trust_region_radius for i in range(N)
        ]

        constraints_initial = [xs[0,0:7] == self.x0[0:7]]
        dg_target = self.target.target_state_derivative(xbar[-1,7])
        constraints_final = [
            xs[-1,0:6] - self.target.target_state(xbar[-1,7]) - dg_target * (xs[-1,7] - xbar[-1,7]) == xis[0:6]
        ]
        
        constraint_t0 = [xs[0,7] == 0.0]
        constraints_tf = [self.tf_bounds[0] <= xs[-1,7],
                          xs[-1,7] <= self.tf_bounds[1]]
        constraints_s = [self.s_bounds[0] <= us[i,3] for i in range(Nseg)] + [us[i,3] <= self.s_bounds[1] for i in range(Nseg)]
        
        constraints_control = [
            vs[i,0] - self.Tmax * np.exp(-xbar[i,6]) * (1 - (xs[i,6] - xbar[i,6])) <= zetas[i]
            for i in range(Nseg)
        ]

        constraints_zeta_positive = [zetas[i] >= 0.0 for i in range(Nseg)]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_objsoc + constraints_dyn + constraints_trustregion +\
            constraints_initial + constraints_final + constraints_control +\
            constraint_t0 + constraints_tf + constraints_s + constraints_zeta_positive)
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return xs.value, us.value, vs.value, None, xis_dyn.value, xis.value, zetas.value
    
    def evaluate_nonlinear_constraints(self, xs, us, vs, ys=None):
        """Evaluate nonlinear constraints
        
        Returns:
            (tuple): tuple of 1D arrays of nonlinear equality and inequality constraints
        """
        g_eq = xs[-1,0:6] - self.target.target_state(xs[-1,7])
        h_ineq = np.array([
            max(vs[i,0] - self.Tmax * np.exp(-xs[i,6]), 0.0) for i in range(self.N-1)
        ])
        return g_eq, h_ineq
    

class FreeTimeContinuousMovingTargetRdvMass(ContinuousControlSCOCP):
    """Free-time continuous moving target rendezvous problem class with mass dynamics
    
    Note the ordering expected for the state and the control vectors: 

    state = [x,y,z,vx,vy,vz,mass,t]
    u = [ux, uy, uz, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    
    The target state is given by a `MovingTarget` object, which result in 6 non-convex equality constraints.
    """
    def __init__(self, x0, target: MovingTarget, Tmax, tf_bounds, s_bounds, N, *args, **kwargs):
        super().__init__(ng=6, nh=0, *args, **kwargs)
        assert len(x0) == 7
        assert abs(self.times[0]  - 0.0) < 1e-14, f"self.times[0] must be 0.0, but given {self.times[0]}"
        assert abs(self.times[-1] - 1.0) < 1e-14, f"self.times[-1] must be 1.0, but given {self.times[-1]}"
        assert s_bounds[0] > 0.0, f"s_bounds[0] must be greater than 0.0, but given {s_bounds[0]}"
        assert s_bounds[0] < s_bounds[1], f"s_bounds[0] must be less than s_bounds[1], but given {s_bounds[0]} and {s_bounds[1]}"
        self.x0 = x0
        self.target = target
        self.Tmax = Tmax
        self.tf_bounds = tf_bounds
        self.s_bounds = s_bounds
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        return -xs[-1,6]
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, gs, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        
        xs      = cp.Variable((N, nx), name='state')
        us      = cp.Variable((Nseg, nu), name='control')
        vs      = cp.Variable((Nseg, 1), name='Gamma')
        xis_dyn = cp.Variable((Nseg,nx), name='xi_dyn')         # slack for dynamics
        xis     = cp.Variable((self.ng,), name='xi')            # slack for target state
        
        penalty = get_augmented_lagrangian_penalty(
            self.weight,
            xis_dyn,
            self.lmb_dynamics,
            xi=xis,
            lmb_eq=self.lmb_eq,
        )
        objective_func = -xs[-1,6] + penalty
        constraints_objsoc = [cp.SOC(vs[i,0], us[i,0:3]) for i in range(N-1)]
        constraints_control = [vs[i,0] <= 1.0 for i in range(Nseg)]

        # constraints on dynamics for state and control
        constraints_dyn = [
            xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,0:4] @ us[i,:] + self.Phi_B[i,:,4] * vs[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
            for i in range(Nseg)
        ]

        # trust region constraints
        constraints_trustregion = [
            xs[i,:] - xbar[i,:] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,:] - xbar[i,:] >= -self.trust_region_radius for i in range(N)
        ]

        # boundary conditions
        constraints_initial = [xs[0,0:7] == self.x0[0:7]]
        dg_target = self.target.target_state_derivative(xbar[-1,7])
        constraints_final = [
            xs[-1,0:6] - self.target.target_state(xbar[-1,7]) - dg_target * (xs[-1,7] - xbar[-1,7]) == xis[0:6]
        ]
        
        # constraints on times
        constraint_t0 = [xs[0,7] == 0.0]
        constraints_tf = [self.tf_bounds[0] <= xs[-1,7],
                          xs[-1,7] <= self.tf_bounds[1]]
        constraints_s = [self.s_bounds[0] <= us[i,3] for i in range(Nseg)] + [us[i,3] <= self.s_bounds[1] for i in range(Nseg)]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_objsoc + constraints_dyn + constraints_trustregion +\
            constraints_initial + constraints_final + constraints_control +\
            constraint_t0 + constraints_tf + constraints_s)
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return xs.value, us.value, vs.value, None, xis_dyn.value, xis.value, None
    
    def evaluate_nonlinear_constraints(self, xs, us, vs, ys=None):
        """Evaluate nonlinear constraints
        
        Returns:
            (tuple): tuple of 1D arrays of nonlinear equality and inequality constraints
        """
        g_eq = xs[-1,0:6] - self.target.target_state(xs[-1,7])
        h_ineq = np.zeros(self.nh)
        return g_eq, h_ineq



"""Test SCP continuous transfer with free final time"""

import cvxpy as cp
import numpy as np
import matplotlib.pyplot as plt

import os
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))

import scocp


def test_freetf_integrator(get_plot=False):
    """Test SCP continuous transfer"""
    mu = 1.215058560962404e-02

    # we create the time-domain integrator for plotting etc.
    integrator_timedomain = scocp.ScipyIntegrator(
        nx=6, nu=3, rhs=scocp.rhs_cr3bp, rhs_stm=scocp.rhs_cr3bp_stm, args=(mu,),
        method='DOP853', reltol=1e-12, abstol=1e-12
    )
    
    # this is the non-dimentional time integrator for solving the OCP
    integrator_01domain = scocp.ScipyIntegrator(
        nx=7,        # state is [x,y,z,vx,vy,vz,t]
        nu=4,        # controls are [ax,ay,az,s]    
        rhs=scocp.control_rhs_cr3bp_freetf, rhs_stm=scocp.control_rhs_cr3bp_freetf_stm,
        impulsive=False,
        args=(mu,[0.0,0.0,0.0,1.0]),   # last argument is dummy placeholder
        method='DOP853', reltol=1e-12, abstol=1e-12
    )

    # propagate uncontrolled and controlled dynamics
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0])
    period_0 = 2.3538670417546639E+00
    sol_lpo0 = integrator_timedomain.solve([0, period_0/2], x0, get_ODESolution=True)

    # integrate with time-dilation factor
    x0_tau = np.concatenate((x0, [0.0]))
    print(f"x0_tau = {x0_tau}")
    sol_lpo_tau = integrator_01domain.solve([0, 1.0], x0_tau,u=np.array([0.0, 0.0, 0.0, period_0/2]),get_ODESolution=True)
    print(f"sol_lpo_tau.y[:,-1] = {sol_lpo_tau.y[:,-1]}")

    assert np.allclose(sol_lpo0.y[0:6,-1], sol_lpo_tau.y[0:6,-1], atol=1e-11)

    if get_plot:
        fig = plt.figure(figsize=(10,5))
        ax = fig.add_subplot(1,2,1,projection='3d')
        ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:], label="lpo0", lw=4.0)
        ax.plot(sol_lpo_tau.y[0,:], sol_lpo_tau.y[1,:], sol_lpo_tau.y[2,:], label="lpo_tau", lw=1.0)
        ax.legend()
        ax.set_aspect('equal')
        
        ax = fig.add_subplot(1,2,2)
        ax.plot(sol_lpo_tau.t, sol_lpo_tau.y[6,:], marker="o", ms=2, color='k')
        ax.axhline(sol_lpo0.t[0], color='r', linestyle='--', label="t0")
        ax.axhline(sol_lpo0.t[-1], color='g', linestyle='--', label="tf")
        ax.set(xlabel="tau", ylabel="time")
        ax.grid(True, alpha=0.5)
    return



def test_scp_scipy_freetf(get_plot=False):
    """Test SCP continuous transfer"""
    mu = 1.215058560962404e-02

    # we create the time-domain integrator for plotting etc.
    integrator_timedomain = scocp.ScipyIntegrator(
        nx=6, nu=3, rhs=scocp.rhs_cr3bp, rhs_stm=scocp.rhs_cr3bp_stm, args=(mu,),
        method='DOP853', reltol=1e-12, abstol=1e-12
    )
    
    # this is the non-dimentional time integrator for solving the OCP
    integrator_01domain = scocp.ScipyIntegrator(
        nx=7,        # state is [x,y,z,vx,vy,vz,t]
        nu=4,        # controls are [ax,ay,az,s]    
        rhs=scocp.control_rhs_cr3bp_freetf,
        rhs_stm=scocp.control_rhs_cr3bp_freetf_stm,
        impulsive=False,
        args=(mu,[0.0,0.0,0.0,1.0]),   # last argument is dummy placeholder
        method='DOP853', reltol=1e-12, abstol=1e-12
    )

    # propagate uncontrolled and controlled dynamics
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0])
    period_0 = 2.3538670417546639E+00
    sol_lpo0 = integrator_timedomain.solve([0, period_0], x0, get_ODESolution=True)

    xf = np.array([
        1.1648780946517576,
        0.0,
        -1.1145303634437023E-1,
        0.0,
        -2.0191923237095796E-1,
        0.0])
    period_f = 3.3031221822879884
    sol_lpo1 = integrator_timedomain.solve([0, period_f], xf, get_ODESolution=True)

    # transfer problem discretization
    N = 40
    tf_bounds = np.array([period_0, period_f])
    tf_guess = np.mean(tf_bounds)
    s_bounds = [0.01*tf_guess, 10*tf_guess]
    
    times_guess = np.linspace(0, tf_guess, N)    # initial guess
    taus = np.linspace(0, 1, N)
    umax = 0.3  # max acceleration

    # create subproblem
    problem = scocp.FreeTimeContinuousRdv(x0, xf, umax, tf_bounds, s_bounds, integrator_01domain, taus)

    # create initial guess
    print(f"Preparing initial guess...")
    sol_initial = integrator_timedomain.solve([0, tf_guess], x0, t_eval=times_guess, get_ODESolution=True)
    sol_final  = integrator_timedomain.solve([0, tf_guess], xf, t_eval=times_guess, get_ODESolution=True)

    alphas = np.linspace(1,0,N)
    xbar = (np.multiply(sol_initial.y, np.tile(alphas, (6,1))) + np.multiply(sol_final.y, np.tile(1-alphas, (6,1)))).T
    xbar[0,:] = x0  # overwrite initial state
    xbar[-1,:] = xf # overwrite final state
    xbar = np.concatenate((xbar,  np.linspace(0, tf_guess, N).reshape(-1,1)), axis=1)
    
    sbar_initial = tf_guess * np.ones((N-1,1))
    ubar = np.concatenate((np.zeros((N-1,3)), sbar_initial), axis=1)
    vbar = np.sum(ubar[:,0:3], axis=1).reshape(-1,1)

    # check initial guess
    _, sols_ig = problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar, steps=5)
    # geq_nl_initial, sols_initial = problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar)
    # fig = plt.figure(figsize=(10,5))
    # ax = fig.add_subplot(1,2,1,projection='3d')
    # for (_ts, _ys) in sols_initial:
    #     ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], 'k-', lw=0.3)
    # ax.scatter(x0[0], x0[1], x0[2], marker='x', color='k', label='Initial state')
    # ax.scatter(xf[0], xf[1], xf[2], marker='o', color='k', label='Final state')
    # ax.scatter(xbar[:,0], xbar[:,1], xbar[:,2], marker='o', color='r', label='Initial guess')
    # ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:], 'k-', lw=0.3)
    # ax.plot(sol_lpo1.y[0,:], sol_lpo1.y[1,:], sol_lpo1.y[2,:], 'k-', lw=0.3)
    # ax.set_aspect('equal')
    # ax = fig.add_subplot(1,2,2)
    # for (_ts,_ys) in sols_initial:
    #     ax.plot(_ts, _ys[:,6], marker="o", ms=2, color='k')
    # ax.grid(True, alpha=0.5)

    # solve subproblem
    problem.solve_convex_problem(xbar, ubar, vbar)
    assert problem.cp_status == "optimal"

    # setup algorithm & solve
    tol_feas = 1e-10
    tol_opt = 1e-4
    algo = scocp.SCvxStar(problem, tol_opt=tol_opt, tol_feas=tol_feas)
    solution = algo.solve(
        xbar,
        ubar,
        vbar,
        maxiter = 100,
        verbose = True
    )
    xopt, uopt, vopt, yopt, sols, summary_dict = solution.x, solution.u, solution.v, solution.y, solution.sols, solution.summary_dict
    assert summary_dict["status"] == "Optimal"
    assert summary_dict["chi"][-1] <= tol_feas
    print(f"Initial guess TOF: {tf_guess:1.4f} --> Optimized TOF: {xopt[-1,6]:1.4f} (bounds: {tf_bounds[0]:1.4f} ~ {tf_bounds[1]:1.4f})")

    # evaluate nonlinear violations
    geq_nl_opt, sols = problem.evaluate_nonlinear_dynamics(xopt, uopt, vopt, steps=20)
    
    # evaluate solution
    if (get_plot is True) and (summary_dict["status"] != "CPFailed"):    
        # plot results
        fig = plt.figure(figsize=(10,7))
        ax = fig.add_subplot(2,3,1,projection='3d')
        for (_ts, _ys) in sols_ig:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], 'b-')

            # interpolate control
            _us_zoh = scocp.zoh_controls(taus, uopt, _ts)
            ax.quiver(_ys[:,0], _ys[:,1], _ys[:,2], _us_zoh[:,0], _us_zoh[:,1], _us_zoh[:,2], color='r', length=0.5)

        ax.scatter(x0[0], x0[1], x0[2], marker='x', color='k', label='Initial state')
        ax.scatter(xf[0], xf[1], xf[2], marker='o', color='k', label='Final state')
        ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:], 'k-', lw=0.3)
        ax.plot(sol_lpo1.y[0,:], sol_lpo1.y[1,:], sol_lpo1.y[2,:], 'k-', lw=0.3)
        ax.set_aspect('equal')
        ax.legend()

        ax_u = fig.add_subplot(2,3,2)
        ax_u.grid(True, alpha=0.5)
        ax_u.step(xopt[:,6], np.concatenate((vopt[:,0], [0.0])), label="Gamma", where='post', color='k')
        ax_u.set(xlabel="Time", ylabel="Control")
        ax_u.legend()

        ax = fig.add_subplot(2,3,3)
        for (_ts, _ys) in sols_ig:
            ax.plot(_ts, _ys[:,6], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ts, _ys[:,6], marker="o", ms=2, color='k')
        ax.grid(True, alpha=0.5)
        ax.set(xlabel="tau", ylabel="time")

        ax_DeltaJ = fig.add_subplot(2,3,4)
        ax_DeltaJ.grid(True, alpha=0.5)
        algo.plot_DeltaJ(ax_DeltaJ, summary_dict)
        ax_DeltaJ.axhline(tol_opt, color='k', linestyle='--', label='tol_opt')
        ax_DeltaJ.legend()

        ax_DeltaL = fig.add_subplot(2,3,5)
        ax_DeltaL.grid(True, alpha=0.5)
        algo.plot_chi(ax_DeltaL, summary_dict)
        ax_DeltaL.axhline(tol_feas, color='k', linestyle='--', label='tol_feas')
        ax_DeltaL.legend()

        plt.tight_layout()
        fig.savefig(os.path.join(os.path.dirname(os.path.abspath(__file__)), "plots/scp_scipy_scipy_freetf.png"), dpi=300)
    return


if __name__ == "__main__":
    #test_freetf_integrator(get_plot=True)
    test_scp_scipy_freetf(get_plot=True)
    plt.show()


"""
Kepler's equation with Der's formulation for computing state-transition matrix, from G. Der (1996). An elegant state transition matrix

Usage:
	statef, stmf = keplerder()
"""

import numpy as np
from numba import njit

from ._misc import get_eccentricity, get_semiMajorAxis


# ------------------------------------------------------------------------------------- #
# hyperbolic trig functions
@njit
def hypertrig_s(z):
    if z > 0.0:
        s = (np.sqrt(z)-np.sin(np.sqrt(z))) / (np.sqrt(z))**3
    elif z < 0.0:
        s = (np.sinh(np.sqrt(-z)) - np.sqrt(-z)) / (np.sqrt(-z))**3
    else:
        s = 1.0/6.0
    return s

@njit
def hypertrig_c(z):
    if z > 0.0:
        c = (1.0 - np.cos(np.sqrt(z)))/z
    elif z < 0.0:
        c = (np.cosh(-z) - 1)/(-z)
    else:
        c = 0.5
    return c


# ------------------------------------------------------------------------------------- #
# Lagrange parameter functions
@njit
def lagrange_us(x, alpha):
    """Function computes U0 ~ U3 from G. Der 1996"""
    # evaluate hypertrig function
    S = hypertrig_s(alpha*x**2)
    C = hypertrig_c(alpha*x**2)
    # parameters
    u0 = 1 - alpha * x**2 * C
    u1 = x*(1 - alpha * x**2 * S)
    u2 = x**2 * C
    u3 = x**3 * S
    return u0, u1, u2, u3


@njit
def kepler_der(x, alpha, t, t0, mu, r0, sigma0):
    """Function computes Kepler's time equation and its derivatives in G. Der 1996 form"""
    u0, u1, u2, u3 = lagrange_us(x, alpha)
    Fun = r0*u1 + sigma0*u2 + u3 - np.sqrt(mu) * (t - t0)
    dF  = r0*u0 + sigma0*u1 + u2
    d2F = sigma0*u0 + (1 - alpha*r0)*u1
    return Fun, dF, d2F


@njit
def lagrange_coefficients_der(mu, x, alpha, r0, v0, sigma0, u0, u1, u2, u3, r, sigma):
    """Function computes Lagrange coefficients (as defined in G. Der 1996)"""
    # evaluate hypertrig function
    S = hypertrig_s(alpha*x**2)
    C = hypertrig_c(alpha*x**2)
    # scalar function
    f = 1.0 - u2/r0
    g = r0*u1 / np.sqrt(mu) + sigma0*u2 / np.sqrt(mu)
    fdot = -np.sqrt(mu) / (r*r0) * u1
    gdot = 1.0 - u2/r
    return f, g, fdot, gdot


# ------------------------------------------------------------------------------------- #
# STM coefficients (eqn. 18)
@njit
def der_stm_coefs(mu, alpha, r0, v0, r, u0, u1, u2, u3, sigma0, sigma):
    """Scalar coefficients of STM dyads"""
    # first row coefficients
    c11 = 1/(alpha*r*r0**2) * (3*u1*u3 + (alpha*r0 - 2)*u2**2) + u1**2/r + u2/r0
    c12 = v0*u1*u2 / (r*np.sqrt(mu))
    c13 = v0/(alpha*r*r0**2*np.sqrt(mu))*(r0*u1*u2 + 2*sigma0*u2**2 + 3*u2*u3 - 3*r*u3 + alpha*r0**2*u1*u2)
    c14 = v0**2 * u2**2 / (r*mu)
    # second row
    c21 = r0*u1*u2 / (r*np.sqrt(mu))
    c22 = v0/(alpha*r*mu) * (3*u1*u3 + (alpha*r0 - 2)*u2**2)
    c23 = r0*v0*u2**2 / (r*mu)
    c24 = v0**2 / (alpha*r*mu*np.sqrt(mu)) * (r0*u1*u2 + 2*sigma0*u2**2 + 3*u2*u3 - 3*r*u3)
    # third row
    c31 = np.sqrt(mu)/(alpha*r**3*r0**2) * (r*(3*u0*u3 - u1*u2 + 2*alpha*r0*u1*u2) - sigma*(3*u1*u3 - 2*u2**2 + alpha*r0*u2**2)) + np.sqrt(mu)/(r**3*r0)*(2*r*r0*u0*u1 + r**2*u1 - sigma*r0*u1**2)
    c32 = v0/r**3 * (r*(u0*u2 + u1**2) - sigma*u1*u2)
    c33 = -3*v0*u2/(alpha*r*r0**2) + v0/(alpha*r**2*r0**2) * (4*sigma0*u1*u2 + r0*(u0*u2 + u1**2) - 3*(u1*u3 + u2**2)) - sigma*v0*u2 / (alpha*r**3*r0**2)*(r0*u1 + 2*sigma0*u2 + 3*u3) + v0/r**3*(r*(u0*u2 + u1**2) - sigma*u1*u2)
    c34 = v0**2 / (r**3*np.sqrt(mu)) * (2*r*u1*u2 - sigma*u2**2)
    # fourth row
    c41 = r0/r**3 * (r*(u0*u2 + u1**2) - sigma*u1*u2)
    c42 = v0/(alpha*r**3*np.sqrt(mu)) * (r*(3*u0*u3 - u1*u2 + 2*alpha*r0*u1*u2) - sigma*(3*u1*u3 - 2*u2**2 + alpha*r0*u2**2))
    c43 = r0*v0/(r**3*np.sqrt(mu)) * (2*r*u1*u2 - sigma*u2**2)
    c44 = -3*v0**2*u2 / (alpha*r*mu) + v0**2 / (alpha*r**2*mu) * (4*sigma0*u1*u2 + r0*(u0*u2 + u1**2) + 3*(u1*u3 + u2**2)) - sigma*v0**2*u2/(alpha*r**3*mu) * (r0*u1 + 2*sigma0*u2 + 3*u3)
    return c11,c12,c13,c14,c21,c22,c23,c24,c31,c32,c33,c34,c41,c42,c43,c44


# ------------------------------------------------------------------------------------- #
# Keplerian time-propagation with G. Der formulation
@njit
def keplerder(mu, state0, t0, t, tol=1e-12, maxiter=10, get_stm=False):
    """Function computes position at some future time t

    Args:
		mu (float): gravitational parameter
		state0 (np.array): initial state [x,y,z,vx,vy,vz]
		t0 (float): initial time
		t (float): final time
		tol (float): tolerance on Laguerre-correction
		maxiter (int): max allowed iteration allowed for Laguerre-correction
        get_stm (bool): whether to get STM; if False, returns (6,6) zeros matrix

    Returns:
		(tuple): final state, final STM
    """
    assert len(state0) == 6, "state0 must be a 6-vector"
    # ------------------------------------------ #
    # SET-UP PROBLEM
    r0, v0 = state0[0:3], state0[3:6]
    sma = get_semiMajorAxis(state0, mu)
    alpha = 1.0/sma
    sigma0 =  np.dot(r0, v0) / np.sqrt(mu)
    
    # ------------------------------------------ #
    # ITERATION WITH LAGUERRE-CONWAY METHOD
    # initial guess
    ecc = np.linalg.norm(get_eccentricity(state0, mu))
    if ecc < 1:
        x0 = alpha*np.sqrt(mu)*(t-t0)
    else:
        x0 = np.sqrt(mu)*(t-t0)/(10*np.linalg.norm(r0))
    # initialize iteratation
    count = 0
    while count < maxiter:
        # evaluate function
        Fun, dF, d2F =  kepler_der(x0, alpha, t, t0, mu, np.linalg.norm(r0), sigma0)
        if np.abs(Fun) < tol:
            break
        # Laguerre-Conway iteration
        x1 = x0 - 5*Fun / (dF + dF/np.abs(dF) * np.sqrt(abs(16*dF**2 - 20*Fun*d2F)) )   
        count += 1
        x0 = x1
    if np.abs(Fun) > tol:
        print("Laguerre-Conway failed! error: ")
    
    # ------------------------------------------ #
    # COMPUTE FINAL POSITION
    # convert back to position
    u0, u1, u2, u3 = lagrange_us(x1, alpha)
    r_scal = np.linalg.norm(r0)*u0 + sigma0*u1 + u2
    sigma = sigma0*u0 + (1-alpha*np.linalg.norm(r0))*u1
    
    # get lagrange coefficients
    f, g, fdot, gdot = lagrange_coefficients_der(mu, x1, alpha, np.linalg.norm(r0), np.linalg.norm(v0), sigma0, u0, u1, u2, u3, r_scal, sigma)
    # create map for state
    fullmap = np.array([[f, 0., 0., g, 0., 0.], [0., f, 0., 0., g, 0.], [0., 0., f, 0., 0., g],
          [fdot, 0., 0., gdot, 0., 0.], [0., fdot, 0., 0., gdot, 0.], [0., 0., fdot, 0., 0., gdot]])
    state1 = np.dot(fullmap, state0)
    
    # ------------------------------------------ #
    # COMPUTE FINAL STM
    if get_stm==True:
        # submatrices M's
        M1 = r0.reshape(3,1) * r0.reshape(1,3) / np.linalg.norm(r0)**2
        M2 = r0.reshape(3,1) * v0.reshape(1,3) / (np.linalg.norm(r0)*np.linalg.norm(v0))
        M3 = v0.reshape(3,1) * r0.reshape(1,3) / (np.linalg.norm(r0)*np.linalg.norm(v0))
        M4 = v0.reshape(3,1) * v0.reshape(1,3) / np.linalg.norm(v0)**2
        # coefficients
        c11,c12,c13,c14,c21,c22,c23,c24,c31,c32,c33,c34,c41,c42,c43,c44 = der_stm_coefs(mu, alpha, np.linalg.norm(r0), 
                                                        np.linalg.norm(v0), r_scal, u0, u1, u2, u3, sigma0, sigma)
        # construct STM
        Rtilda = f*np.identity(3)    + c11*M1 + c12*M2 + c13*M3 + c14*M4
        R      = g*np.identity(3)    + c21*M1 + c22*M2 + c23*M3 + c24*M4
        Vtilda = fdot*np.identity(3) + c31*M1 + c32*M2 + c33*M3 + c34*M4
        V      = gdot*np.identity(3) + c41*M1 + c42*M2 + c43*M3 + c44*M4
        stm = np.concatenate(( np.concatenate((Rtilda,R), axis=1), np.concatenate((Vtilda,V), axis=1) ), axis=0 )
    else:
        stm = np.zeros((6,6))
    return state1, stm


# Keplerian time-propagation with G. Der formulation, with no STM
@njit
def keplerder_nostm(mu, state0, t0, t, tol=1e-12, maxiter=10):
    """Function computes position at some future time t

    Args:
        mu (float): gravitational parameter
        state0 (np.array): initial state [x,y,z,vx,vy,vz]
        t0 (float): initial time
        t (float): final time
        tol (float): tolerance on Laguerre-correction
        maxiter (int): max allowed iteration allowed for Laguerre-correction

    Returns:
        (np.array): final state
    """
    assert len(state0) == 6, "state0 must be a 6-vector"
    # SET-UP PROBLEM
    sqrt_mu = np.sqrt(mu)
    r0, v0 = state0[0:3], state0[3:6]
    sma = get_semiMajorAxis(state0, mu)
    alpha = 1.0/sma
    sigma0 =  (r0[0]*v0[0] + r0[1]*v0[1] + r0[2]*v0[2])/sqrt_mu 
    # np.dot(r0, v0) / np.sqrt(mu)
    
    # ------------------------------------------ #
    # ITERATION WITH LAGUERRE-CONWAY METHOD
    # initial guess
    ecc = np.linalg.norm(get_eccentricity(state0, mu))
    if ecc < 1:
        x0 = alpha*sqrt_mu*(t-t0)
    else:
        x0 = sqrt_mu*(t-t0)/(10*np.linalg.norm(r0))
    # initialize iteratation
    count = 0
    while count < maxiter:
        # evaluate function
        Fun, dF, d2F =  kepler_der(x0, alpha, t, t0, mu, np.linalg.norm(r0), sigma0)
        if np.abs(Fun) < tol:
            break
        # Laguerre-Conway iteration
        x1 = x0 - 5*Fun / (dF + dF/np.abs(dF) * np.sqrt(abs(16*dF**2 - 20*Fun*d2F)) )   
        count += 1
        x0 = x1
    if np.abs(Fun) > tol:
        print("Laguerre-Conway failed! error: ")
    
    # ------------------------------------------ #
    # COMPUTE FINAL POSITION
    # convert back to position
    u0, u1, u2, u3 = lagrange_us(x1, alpha)
    r_scal = np.linalg.norm(r0)*u0 + sigma0*u1 + u2
    sigma = sigma0*u0 + (1-alpha*np.linalg.norm(r0))*u1
    
    # get lagrange coefficients
    f, g, fdot, gdot = lagrange_coefficients_der(mu, x1, alpha, np.linalg.norm(r0), np.linalg.norm(v0), sigma0, u0, u1, u2, u3, r_scal, sigma)
    # create map for state
    state1 = np.array([
            f*state0[0] + state0[3]*g, 
            f*state0[1] + state0[4]*g,
            f*state0[2] + state0[5]*g,
            fdot*state0[0] + gdot*state0[3], 
            fdot*state0[1] + gdot*state0[4], 
            fdot*state0[2] + gdot*state0[5], 
        ])
    return state1


"""Test rendezvous with moving target"""

import cvxpy as cp
import matplotlib.pyplot as plt
import numpy as np

import os
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))

import scocp


def test_scp_scipy_logmass(get_plot=False):
    """Test SCP continuous transfer with log-mass dynamics"""
    # define canonical parameters
    GM_SUN = 132712000000.44     # Sun GM, km^3/s^-2
    MSTAR  = 1000.0              # reference spacecraft mass
    ISP    = 3000.0              # specific impulse, s
    THRUST = 0.3                 # max thrust, kg.m/s^2
    G0     = 9.81                # gravity at surface, m/s^2

    DU = 149.6e6                 # length scale set to Sun-Earth distance, km
    VU = np.sqrt(GM_SUN / DU)    # velocity scale, km/s
    TU = DU / VU                 # time scale, s

    isp = ISP/TU                                # canonical specific impulse, TU
    Tmax = THRUST * (1/MSTAR)*(TU**2/(1e3*DU))  # canonical max thrust
    cex = isp * G0*(TU**2/(1e3*DU))             # canonical exhaust velocity

    mu = 1.0
    parameters_ode = (mu, cex)

    print(f"\nCanonical isp: {isp:1.4e} TU, cex: {cex:1.4e} DU/TU, tmax: {Tmax:1.4e} MU.DU/TU^2")

    # we create the time-domain integrator for plotting etc.
    integrator_timedomain = scocp.ScipyIntegrator(
        nx=6,
        nu=0,
        rhs=scocp.rhs_twobody,
        rhs_stm=scocp.rhs_twobody_stm,
        args=(mu,),
        method='DOP853',
        reltol=1e-12,
        abstol=1e-12
    )
    
    # this is the non-dimentional time integrator for solving the OCP
    integrator_01domain = scocp.ScipyIntegrator(
        nx=8,
        nu=4,
        nv=1,
        rhs=scocp.control_rhs_twobody_logmass_freetf,
        rhs_stm=scocp.control_rhs_twobody_logmass_freetf_stm,
        impulsive=False,
        args=(parameters_ode,[0.0,0.0,0.0,1.0,0.0]),
        method='DOP853', reltol=1e-12, abstol=1e-12)
    
    # propagate uncontrolled and controlled dynamics
    oe0 = np.array([1., 0.0, np.deg2rad(1), 1.5, 0.0, np.deg2rad(0)])
    x0 = np.concatenate((scocp.kep2rv(oe0, mu), [np.log(1.0)]))
    period_0 = 2*np.pi*np.sqrt(oe0[0]**3/mu)
    sol_lpo0 = integrator_timedomain.solve([0, period_0], x0[0:6], get_ODESolution=True)

    oef0 = np.array([1.5, 0.01, np.deg2rad(3), 0.3, 0.0, np.deg2rad(130)])
    xf0 = scocp.kep2rv(oef0, mu)
    period_f = 2*np.pi*np.sqrt(oef0[0]**3/mu)
    sol_lpo1 = integrator_timedomain.solve([0, period_f], xf0[0:6], get_ODESolution=True)

    # transfer problem discretization
    N = 40
    tf_bounds = np.array([period_0, 1.3 * period_f])
    tf_guess = 2*np.pi
    s_bounds = [0.01*tf_guess, 10*tf_guess]
    
    times_guess = np.linspace(0, tf_guess, N)    # initial guess
    taus = np.linspace(0, 1, N)

    # create target object
    def eval_target_state(t: float) -> np.ndarray:
        return scocp.keplerder_nostm(mu, xf0, 0.0, t, tol=1e-12, maxiter=10)
    
    def eval_target_state_derivative(t: float) -> np.ndarray:
        state = eval_target_state(t)
        return scocp.rhs_twobody(t, state, mu)
    
    target = scocp.MovingTarget(eval_target_state, eval_target_state_derivative)

    # evaluate target state
    xf_guess = target.eval_state(tf_guess)

    # create subproblem
    problem = scocp.FreeTimeContinuousMovingTargetRdvLogMass(
        x0, target, Tmax, tf_bounds, s_bounds, N, integrator_01domain, taus, augment_Gamma=True,
        weight = 10.0
    )

    # create initial guess
    print(f"Preparing initial guess...")
    elements = np.concatenate((
        np.linspace(oe0[0], oef0[0], N).reshape(-1,1),
        np.linspace(oe0[1], oef0[1], N).reshape(-1,1),
        np.linspace(oe0[2], oef0[2], N).reshape(-1,1),
        np.linspace(oe0[3], oef0[3], N).reshape(-1,1),
        np.linspace(oe0[4], oef0[4], N).reshape(-1,1),
        np.linspace(oe0[5], oef0[5], N).reshape(-1,1),
    ), axis=1)
    elements[:,5] = np.linspace(oe0[5], oef0[5]+2*np.pi, N)
    xbar = np.zeros((N,8))
    xbar[:,0:6] = np.array([scocp.kep2rv(E,mu) for E in elements])
    xbar[:,6]   = np.log(np.linspace(1.0, 0.5, N))  # initial guess for log-mass
    xbar[0,0:7]   = x0[0:7]          # overwrite initial state
    xbar[-1,0:6]  = xf_guess[0:6]          # overwrite final state
    xbar[:,7]   = times_guess        # initial guess for time

    sbar_initial = tf_guess * np.ones((N-1,1))
    # ubar = np.concatenate((np.divide(np.diff(xbar[:,3:6], axis=0), np.diff(times_guess)[:,None]), sbar_initial), axis=1)
    ubar = np.concatenate((np.zeros((N-1,3)), sbar_initial), axis=1)
    vbar = np.sum(ubar[:,0:3], axis=1).reshape(-1,1)
    geq_nl_ig, sols_ig = problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar, steps=5) # evaluate initial guess
    
    # solve subproblem
    print(f"ubar.shape = {ubar.shape}, xbar.shape = {xbar.shape}, vbar.shape = {vbar.shape}")
    print(f"problem.Phi_A.shape = {problem.Phi_A.shape}, problem.Phi_B.shape = {problem.Phi_B.shape}, problem.Phi_c.shape = {problem.Phi_c.shape}")
    problem.solve_convex_problem(xbar, ubar, vbar)
    assert problem.cp_status == "optimal", f"CP status: {problem.cp_status}"

    # setup algorithm & solve
    tol_feas = 1e-8
    tol_opt = 1e-6
    algo = scocp.SCvxStar(problem, tol_opt=tol_opt, tol_feas=tol_feas)
    solution = algo.solve(
        xbar,
        ubar,
        vbar,
        maxiter = 100,
        verbose = True
    )
    xopt, uopt, vopt, yopt, sols, summary_dict = solution.x, solution.u, solution.v, solution.y, solution.sols, solution.summary_dict
    assert summary_dict["status"] == "Optimal"
    assert summary_dict["chi"][-1] <= tol_feas
    print(f"Initial guess TOF: {tf_guess:1.4f} --> Optimized TOF: {xopt[-1,7]:1.4f} (bounds: {tf_bounds[0]:1.4f} ~ {tf_bounds[1]:1.4f})")
    xf = target.eval_state(xopt[-1,7])

    # trajectory summary
    print(f"  Departure time : {xopt[0,7]*TU/86400.0:1.4f} days")
    print(f"  Arrival time   : {xopt[-1,7]*TU/86400.0:1.4f} days")
    print(f"  TOF            : {(xopt[-1,7] - xopt[0,7])*TU/86400.0:1.4f} days")
    print(f"  Initial mass   : {np.exp(xopt[0,6])*MSTAR:1.4f} kg")
    print(f"  Final mass     : {np.exp(xopt[-1,6])*MSTAR:1.4f} kg")

    # evaluate nonlinear violations
    geq_nl_opt, sols = problem.evaluate_nonlinear_dynamics(xopt, uopt, vopt, steps=5)
    assert np.max(np.abs(geq_nl_opt)) <= tol_feas
    
    # evaluate solution
    if (get_plot is True) and (summary_dict["status"] != "CPFailed"):
        # plot results
        fig = plt.figure(figsize=(12,7))
        ax = fig.add_subplot(2,3,1,projection='3d')
        ax.set(xlabel="x", ylabel="y", zlabel="z")
        for (_ts, _ys) in sols_ig:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], 'b-')

            # interpolate control
            _us_zoh = scocp.zoh_controls(taus, uopt, _ts)
            ax.quiver(_ys[:,0], _ys[:,1], _ys[:,2], _us_zoh[:,0], _us_zoh[:,1], _us_zoh[:,2], color='r', length=1.0)

        ax.scatter(x0[0], x0[1], x0[2], marker='x', color='k', label='Initial state')
        ax.scatter(xf_guess[0], xf_guess[1], xf_guess[2], marker='^', color='k', label='Final state (guess)')
        ax.scatter(xf[0], xf[1], xf[2], marker='o', color='k', label='Final state')
        ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:], 'k-', lw=0.3)
        ax.plot(sol_lpo1.y[0,:], sol_lpo1.y[1,:], sol_lpo1.y[2,:], 'k-', lw=0.3)
        #ax.set_aspect('equal')
        ax.legend()

        ax_m = fig.add_subplot(2,3,2)
        ax_m.grid(True, alpha=0.5)
        for (_ts, _ys) in sols:
            ax_m.plot(_ys[:,7], np.exp(_ys[:,6]), 'b-')
        ax_m.axhline(np.exp(sols[-1][1][-1,6]), color='r', linestyle='--')
        ax_m.text(0.0, 0.01 + np.exp(sols[-1][1][-1,6]), f"m_f = {np.exp(sols[-1][1][-1,6]):1.4f}", color='r')
        ax_m.set(xlabel="Time", ylabel="Mass")
        ax_m.legend()

        ax_u = fig.add_subplot(2,3,3)
        ax_u.grid(True, alpha=0.5)
        ax_u.step(xopt[:,7], np.concatenate((vopt[:,0], [0.0])), label="Control", where='post', color='k')
        ax_u.step(xopt[:,7], np.concatenate((uopt[:,0], [0.0])), label="ux", where='post', color='r')
        ax_u.step(xopt[:,7], np.concatenate((uopt[:,1], [0.0])), label="uy", where='post', color='b')
        ax_u.step(xopt[:,7], np.concatenate((uopt[:,2], [0.0])), label="uz", where='post', color='g')
        for idx, (_ts, _ys) in enumerate(sols):
            ax_u.plot(_ys[:,7], Tmax/np.exp(_ys[:,6]), color='r', linestyle=':', label="Max accel." if idx == 0 else None)
        ax_u.set(xlabel="Time", ylabel="Acceleration")
        ax_u.legend()

        ax_DeltaJ = fig.add_subplot(2,3,4)
        ax_DeltaJ.grid(True, alpha=0.5)
        algo.plot_DeltaJ(ax_DeltaJ, summary_dict)
        ax_DeltaJ.axhline(tol_opt, color='k', linestyle='--', label='tol_opt')
        ax_DeltaJ.legend()

        ax_DeltaL = fig.add_subplot(2,3,5)
        ax_DeltaL.grid(True, alpha=0.5)
        algo.plot_chi(ax_DeltaL, summary_dict)
        ax_DeltaL.axhline(tol_feas, color='k', linestyle='--', label='tol_feas')
        ax_DeltaL.legend()

        ax = fig.add_subplot(2,3,6)
        for (_ts, _ys) in sols_ig:
            ax.plot(_ts, _ys[:,7], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ts, _ys[:,7], marker="o", ms=2, color='k')
        ax.grid(True, alpha=0.5)
        ax.set(xlabel="tau", ylabel="time")

        plt.tight_layout()
        fig.savefig(os.path.join(os.path.dirname(os.path.abspath(__file__)), "plots/twobody_logmass_freetf_rdv.png"), dpi=300)
    return


if __name__ == "__main__":
    test_scp_scipy_logmass(get_plot=True)
    plt.show()


"""scocp: Sequential Convex Optimization Control Problem"""

__copyright__    = 'Copyright (C) 2025 Yuri Shimane'
__version__      = '0.1.7'
__license__      = 'GPL-3.0 License'
__author__       = 'Yuri Shimane'
__author_email__ = 'yuri.shimane@gatech.edu'
__url__          = 'https://github.com/Yuricst/scocp'


# check for dependencies
_hard_dependencies = ("cvxpy", "numba", "numpy", "matplotlib", "scipy")
_missing_dependencies = []
for _dependency in _hard_dependencies:
    try:
        __import__(_dependency)
    except ImportError as _e:  # pragma: no cover
        _missing_dependencies.append(f"{_dependency}: {_e}")

if _missing_dependencies:  # pragma: no cover
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(_missing_dependencies)
    )
del _hard_dependencies, _dependency, _missing_dependencies

# miscellaneous functions
from ._misc import zoh_control, zoh_controls, get_augmented_lagrangian_penalty, MovingTarget, kep2rv, rv2kep, rv2mee, mee2rv
from ._keplerder import keplerder, keplerder_nostm

# functions for integrating dynamics
from .eoms import *
from ._integrator_scipy import ScipyIntegrator

# sequentially convexified optimal control problems
from .scocp_impulsive import *
from .scocp_continuous import *

# SCP algorithm
from ._scvxstar import SCvxStar



"""Miscellaneous functions"""

from collections.abc import Callable
import cvxpy as cp
import numpy as np
from numba import njit

def zoh_control(times, us, t):
    """Zero-order hold control"""
    for i in range(len(times)-1):
        # Binary search to find interval containing t
        idx = np.searchsorted(times, t, side='right') - 1
        if idx >= 0 and idx < len(times)-1:
            return us[idx]
    return us[-1]  # Return last control if t > times[-1]


def zoh_controls(times, us, t_eval):
    """Zero-order hold control"""
    _,nu = us.shape
    us_zoh = np.zeros((len(t_eval),nu))
    for i,t in enumerate(t_eval):
        us_zoh[i,:] = zoh_control(times, us, t)
    return us_zoh


def get_augmented_lagrangian_penalty(weight, xi_dyn, lmb_dyn, xi=None, lmb_eq=None, zeta=None, lmb_ineq=None):
    """Evaluate augmented Lagrangian penalty function
    
    Args:
        weight (float): weight of the penalty function
        xi_dyn (cp.Variable): slack variable for dynamics
        lmb_dyn (cp.Parameter): multiplier for dynamics
        xi (cp.Variable, optional): slack variable for equality constraints
        lmb_eq (cp.Parameter, optional): multiplier for equality constraints
        zeta (cp.Variable, optional): slack variable for inequality constraints
        lmb_ineq (cp.Parameter, optional): multiplier for inequality constraints
    
    Returns:
        (cp.Expression): augmented Lagrangian penalty function
    """
    assert xi_dyn.shape == lmb_dyn.shape, f"xi_dyn.shape = {xi_dyn.shape} must match lmb_dyn.shape = {lmb_dyn.shape}"
    penalty = weight/2 * cp.sum_squares(xi_dyn)
    for i in range(lmb_dyn.shape[0]):
        penalty += lmb_dyn[i,:] @ xi_dyn[i,:]
    if xi is not None:
        penalty += weight/2 * cp.sum_squares(xi)
        for i,_ in enumerate(lmb_eq):
            penalty += lmb_eq[i] * xi[i]
    if zeta is not None:                # Note: we do not need to clip zeta as we assume the variable is defined to be >=0 
        penalty += weight/2 * cp.sum_squares(zeta)
        for i,_ in enumerate(lmb_ineq):
            penalty += lmb_ineq[i] * zeta[i]
    return penalty


class MovingTarget:
    """Define moving target for rendezvous problem
    
    We assume the target state equality constraint is of the form

    g(r_N,v_N,t_N) = [ r_N - r_ref(t_N)
                       v_N - v_ref(t_N) ]

    where r_ref(t_N) and v_ref(t_N) are the position and velocity of the target at time t_N.

    Args:
        eval_state (function): function to evaluate target state
        eval_state_derivative (function): function to evaluate derivative of target state
    """
    def __init__(self, eval_state: Callable, eval_state_derivative: Callable):
        self.eval_state = eval_state
        self.eval_state_derivative = eval_state_derivative
        return
    
    def target_state(self, t: float) -> np.ndarray:
        """Get target state at time t"""
        return self.eval_state(t)
    
    def target_state_derivative(self, t: float) -> np.ndarray:
        """Get target state derivative w.r.t. time"""
        return self.eval_state_derivative(t)
    

# --------------------- Function to convert between Keplerian elements and Cartesian state --------------------- #
# See: Curtis pg.209
@njit
def _rotmat_ax1(phi):
    return np.array([ [1.0, 0.0,          0.0], 
                      [0.0,  np.cos(phi), np.sin(phi)], 
                      [0.0, -np.sin(phi), np.cos(phi)] ])

@njit
def _rotmat_ax2(phi):
    return np.array([ [np.cos(phi), 0.0, -np.sin(phi)], 
                      [0.0,         1.0,  0.0        ], 
                      [np.sin(phi), 0.0,  np.cos(phi)] ])

@njit
def _rotmat_ax3(phi):
    return np.array([ [ np.cos(phi), np.sin(phi), 0.0], 
                      [-np.sin(phi), np.cos(phi), 0.0], 
                      [0.0,          0.0,         1.0] ])


@njit
def kep2rv(elements, mu):
    """Get Cartesian state-vector to Keplerian elements
	
    Args:
        elements (np.array): array of Keplerian elements, [sma, ecc, inc, raan, aop, ta]
        mu (float): gravitational parameter

    Returns:
        (np.array): state-vector in Cartesian frame
    """
    # unpack elements
    sma, ecc, inc, raan, aop, ta = elements
    # angular momentum vector
    h   = np.sqrt(sma*mu*(1-ecc**2))
    # perifocal vector
    x   = (h**2/mu) * (1/(1 + ecc*np.cos(ta))) * np.cos(ta)
    y   = (h**2/mu) * (1/(1 + ecc*np.cos(ta))) * np.sin(ta)
    vx  = (mu/h)*(-np.sin(ta))
    vy  = (mu/h)*(ecc + np.cos(ta))
    rpf = np.array([ x, y, 0.0 ])
    vpf = np.array([ vx, vy, 0.0 ])

    # rotate by aop
    r_rot1 = np.dot(_rotmat_ax3(-aop), rpf)
    v_rot1 = np.dot(_rotmat_ax3(-aop), vpf)
    # rotate by inclination
    r_rot2 = np.dot(_rotmat_ax1(-inc), r_rot1)
    v_rot2 = np.dot(_rotmat_ax1(-inc), v_rot1)
    # rotate by raan
    r_rot3 = np.dot(_rotmat_ax3(-raan), r_rot2)
    v_rot3 = np.dot(_rotmat_ax3(-raan), v_rot2)

    # save inertial state
    state_inr = np.concatenate((r_rot3, v_rot3), axis=0) 
    return state_inr


@njit
def get_inclination(state):
    """Function computes inclination in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame

    Returns:
        (float): inclination in radians
    """

    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.cross(r,v)
    # inclination
    inc = np.arccos(h[2] / np.linalg.norm(h))
    return inc


@njit
def get_raan(state):
    """Function computes RAAN in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame

    Returns:
        (float): RAAN in radians
    """
    
    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    zdir = np.array([0, 0, 1])
    ndir = np.cross(zdir, h)
    # compute RAAN
    raan = np.arctan2(ndir[1], ndir[0])
    return raan
    
    
@njit
def get_eccentricity(state, mu):
    """Function computes eccentricity vector from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter

    Returns:
        (np.arr): eccentricity vector
    """
    
    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    #zdir = np.array([0, 0, 1])
    #ndir = np.cross(zdir, h)
    # compute eccentricity vector
    ecc = (1/mu) * np.cross(v,h) - r/np.linalg.norm(r)
    return ecc


@njit
def get_omega(state, mu):
    """Function computes argument of periapsis in radians from a two-body state vector, in inertially frozen frame.
    If eccentricity is 0, omega = RAAN
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter

    Returns:
        (float): argument of periapsis in radians
    """
    
    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.cross(r,v)
    # normal direction of xy-plane
    zdir = np.array([0, 0, 1])
    ndir = np.cross(zdir, h)
    # compute eccentricity vector
    ecc = (1/mu) * np.cross(v,h) - r/np.linalg.norm(r)
    if np.linalg.norm(ndir)*np.linalg.norm(ecc)!=0:
        # compute argument of periapsis
        omega = np.arccos( np.dot(ndir,ecc) / (np.linalg.norm(ndir)*np.linalg.norm(ecc)) )
        if ecc[2] < 0:
            omega = 2*np.pi - omega
    else:
        omega = get_raan(state)
    return omega


@njit
def get_trueanomaly(state, mu):
    """Function computes argument of periapsis in radians from a two-body state vector, in inertially frozen frame
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter

    Returns:
        (float): true anomaly in radians
    """
    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.linalg.norm( np.cross(r,v) )
    # radial velocity
    vr = np.dot(v,r)/np.linalg.norm(r)
    theta = np.arctan2(h*vr, h**2/np.linalg.norm(r) - mu)
    return theta


@njit
def get_semiMajorAxis(state, mu):
    """Function computes semi major axis of keplrian orbit
    
    Args:
        state (np.array): array of cartesian state in inertially frozen frame
        mu (float): two-body mass parameter

    Returns:
        (float): semi-major axis
    """
    # decompose state to position and velocity vector
    r = np.array([state[0], state[1], state[2]])
    v = np.array([state[3], state[4], state[5]])
    # angular momentum
    h = np.linalg.norm( np.cross(r,v) )
    # eccentricity
    e = np.linalg.norm( get_eccentricity(state, mu) )
    # semi-major axis
    a = h**2 / (mu*(1 - e**2))
    return a


@njit
def rv2kep(state, mu, fictious_vz=1.e-15):
    """Get Keplerian elements from Cartesian state-vector and gravitational parameter

    Args:
        state (np.array): state-vector
        mu (float): gravitational parameter
        fictious_vz (float): fictious vz to inject if state is purely planar

    Returns:
        (tuple): eccentricity, semi-major axis, inclination, raan, argument of periapsis, true anomaly
    """
    # if state is planar, inject fictious out-of-plane component
    if (state[2] == 0.) and (state[5] == 0.):
        state[5] = fictious_vz
    ecc  = np.linalg.norm(get_eccentricity(state, mu))
    sma  = get_semiMajorAxis(state, mu)
    inc  = get_inclination(state)
    raan = get_raan(state)
    aop  = get_omega(state, mu)
    ta   = get_trueanomaly(state, mu)
    return np.array([sma, ecc, inc, raan, aop, ta])


def rv2mee(state, mu):
    """Convert Cartesian state to mean elements
    
    Args:
        state (np.array): Cartesian state
        mu (float): gravitational parameter

    Returns:
        (np.array): mean elements
    """
    a,e,i,W,w,ta = rv2kep(state, mu)
    p = a*(1-e**2)
    f = e*np.cos(w+W)
    g = e*np.sin(w+W)
    h = np.tan(i/2) * np.cos(W)
    k = np.tan(i/2) * np.sin(W)
    L = W + w + ta
    return np.array([p,f,g,h,k,L])
    

def mee2rv(mee, mu):
    """Convert mean elements to Cartesian state
    
    Args:
        mee (np.array): mean elements
        mu (float): gravitational parameter

    Returns:
        (np.array): Cartesian state
    """
    p,f,g,h,k,L = mee
    alpha2 = h**2 - k**2
    s2 = 1 + h**2 + k**2
    cosL = np.cos(L)
    sinL = np.sin(L)
    w = 1 + f*cosL + g*sinL
    r = p/w
    sqrt_mu_p = np.sqrt(mu/p)

    rv = np.array([
        r/s2 * (cosL + alpha2*cosL + 2*h*k*sinL),
        r/s2 * (sinL - alpha2*sinL + 2*h*k*cosL),
        2*r/s2 * (h*sinL - k*cosL),
        -1/s2 * sqrt_mu_p * ( sinL + alpha2*sinL - 2*h*k*cosL + g - 2*g*h*k + alpha2*g),
        -1/s2 * sqrt_mu_p * (-cosL + alpha2*cosL + 2*h*k*sinL - f + 2*g*h*k + alpha2*f),
         2/s2 * sqrt_mu_p * (h*cosL + k*sinL + f*h + g*k)
    ])
    return rv



"""Test SCP continuous transfer with log-mass dynamics

For OCP with mass dynamics, make sure to do the following:
- Set `impulsive = False` in the integrator
- Set `augment_Gamma = True` in the problem class
"""

import cvxpy as cp
import numpy as np
import matplotlib.pyplot as plt

import os
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))

import scocp


def test_scp_scipy_logmass(get_plot=False):
    """Test SCP continuous transfer with log-mass dynamics"""
    mu = 1.215058560962404e-02
    cex = 0.8       # exhaust velocity
    parameters_ode = (mu, cex)

    # we create the time-domain integrator for plotting etc.
    integrator_timedomain = scocp.ScipyIntegrator(
        nx=7,
        nu=3,
        rhs=scocp.control_rhs_cr3bp_logmass,
        rhs_stm=scocp.control_rhs_cr3bp_logmass_stm,
        args=(parameters_ode,[0.0,0.0,0.0,0.0]),
        method='DOP853',
        reltol=1e-12,
        abstol=1e-12
    )
    
    # this is the non-dimentional time integrator for solving the OCP
    integrator_01domain = scocp.ScipyIntegrator(
        nx=8,
        nu=4,
        nv=1,
        rhs=scocp.control_rhs_cr3bp_logmass_freetf,
        rhs_stm=scocp.control_rhs_cr3bp_logmass_freetf_stm,
        impulsive=False,
        args=(parameters_ode,[0.0,0.0,0.0,1.0,0.0]),
        method='DOP853', reltol=1e-12, abstol=1e-12)
    
    # propagate uncontrolled and controlled dynamics
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0,
        np.log(1.0),                        # initial log-mass (fixed)
    ])
    period_0 = 2.3538670417546639E+00
    sol_lpo0 = integrator_timedomain.solve([0, period_0], x0, get_ODESolution=True)

    xf = np.array([
        1.1648780946517576,
        0.0,
        -1.1145303634437023E-1,
        0.0,
        -2.0191923237095796E-1,
        0.0,
        np.log(0.5),                        # final log-mass (guess)
    ])
    period_f = 3.3031221822879884
    sol_lpo1 = integrator_timedomain.solve([0, period_f], xf, get_ODESolution=True)

    # transfer problem discretization
    N = 40
    tf_bounds = np.array([period_0, 1.3 * period_f])
    tf_guess = period_f
    s_bounds = [0.01*tf_guess, 10*tf_guess]
    
    times_guess = np.linspace(0, tf_guess, N)    # initial guess
    taus = np.linspace(0, 1, N)
    Tmax = 0.37637494800142673      # max thrust

    # create subproblem
    problem = scocp.FreeTimeContinuousRdvLogMass(
        x0, xf, Tmax, tf_bounds, s_bounds, N, integrator_01domain, taus, augment_Gamma=True,
        weight = 10.0
    )

    # create initial guess
    print(f"Preparing initial guess...")
    sol_initial = integrator_timedomain.solve([0, times_guess[-1]], x0, t_eval=times_guess, get_ODESolution=True)
    sol_final  = integrator_timedomain.solve([0, times_guess[-1]], xf, t_eval=times_guess, get_ODESolution=True)

    alphas = np.linspace(1,0,N)
    xbar = (np.multiply(sol_initial.y, np.tile(alphas, (7,1))) + np.multiply(sol_final.y, np.tile(1-alphas, (7,1)))).T
    xbar[:,6]   = np.log(np.linspace(1.0, 0.5, N))  # initial guess for log-mass
    xbar[0,:]   = x0                 # overwrite initial state
    xbar[-1,:]  = xf                 # overwrite final state
    xbar = np.concatenate((xbar,  times_guess.reshape(-1,1)), axis=1)    # append initial guess for time

    sbar_initial = tf_guess * np.ones((N-1,1))
    ubar = np.concatenate((np.divide(np.diff(xbar[:,3:6], axis=0), np.diff(times_guess)[:,None]), sbar_initial), axis=1)
    vbar = np.sum(ubar[:,0:3], axis=1).reshape(-1,1)
    print(f"Initial guess objective: {problem.evaluate_objective(xbar, ubar, vbar):1.4e}")

    geq_nl_ig, sols_ig = problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar, steps=5)
    print(f"np.linalg.norm(geq_nl_ig) = {np.linalg.norm(geq_nl_ig)}")
    
    # solve subproblem
    print(f"ubar.shape = {ubar.shape}, xbar.shape = {xbar.shape}, vbar.shape = {vbar.shape}")
    print(f"problem.Phi_A.shape = {problem.Phi_A.shape}, problem.Phi_B.shape = {problem.Phi_B.shape}, problem.Phi_c.shape = {problem.Phi_c.shape}")
    problem.solve_convex_problem(xbar, ubar, vbar)
    assert problem.cp_status == "optimal"

    # setup algorithm & solve
    tol_feas = 1e-8
    tol_opt = 1e-6
    algo = scocp.SCvxStar(problem, tol_opt=tol_opt, tol_feas=tol_feas)
    solution = algo.solve(
        xbar,
        ubar,
        vbar,
        maxiter = 250,
        verbose = True
    )
    xopt, uopt, vopt, yopt, sols, summary_dict = solution.x, solution.u, solution.v, solution.y, solution.sols, solution.summary_dict
    assert summary_dict["status"] == "Optimal"
    assert summary_dict["chi"][-1] <= tol_feas
    print(f"Initial guess TOF: {tf_guess:1.4f} --> Optimized TOF: {xopt[-1,7]:1.4f} (bounds: {tf_bounds[0]:1.4f} ~ {tf_bounds[1]:1.4f})")
    
    # evaluate nonlinear violations
    geq_nl_opt, sols = problem.evaluate_nonlinear_dynamics(xopt, uopt, vopt, steps=20)
    assert np.max(np.abs(geq_nl_opt)) <= tol_feas
    
    # evaluate solution
    if (get_plot is True) and (summary_dict["status"] != "CPFailed"):
        # plot results
        fig = plt.figure(figsize=(12,7))
        ax = fig.add_subplot(2,3,1,projection='3d')
        for (_ts, _ys) in sols_ig:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], 'b-')

            # interpolate control
            _us_zoh = scocp.zoh_controls(taus, uopt, _ts)
            ax.quiver(_ys[:,0], _ys[:,1], _ys[:,2], _us_zoh[:,0], _us_zoh[:,1], _us_zoh[:,2], color='r', length=0.1)

        ax.scatter(x0[0], x0[1], x0[2], marker='x', color='k', label='Initial state')
        ax.scatter(xf[0], xf[1], xf[2], marker='o', color='k', label='Final state')
        ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:], 'k-', lw=0.3)
        ax.plot(sol_lpo1.y[0,:], sol_lpo1.y[1,:], sol_lpo1.y[2,:], 'k-', lw=0.3)
        ax.set_aspect('equal')
        ax.legend()

        ax_m = fig.add_subplot(2,3,2)
        ax_m.grid(True, alpha=0.5)
        for (_ts, _ys) in sols:
            ax_m.plot(_ys[:,7], np.exp(_ys[:,6]), 'b-')
        ax_m.axhline(np.exp(sols[-1][1][-1,6]), color='r', linestyle='--')
        ax_m.text(0.0, 0.01 + np.exp(sols[-1][1][-1,6]), f"m_f = {np.exp(sols[-1][1][-1,6]):1.4f}", color='r')
        ax_m.set(xlabel="Time", ylabel="Mass")
        ax_m.legend()

        ax_u = fig.add_subplot(2,3,3)
        ax_u.grid(True, alpha=0.5)
        ax_u.step(xopt[:,7], np.concatenate((vopt[:,0], [0.0])), label="Control", where='post', color='k')
        for idx, (_ts, _ys) in enumerate(sols):
            ax_u.plot(_ys[:,7], Tmax/np.exp(_ys[:,6]), color='r', linestyle=':', label="Max accel." if idx == 0 else None)
        ax_u.set(xlabel="Time", ylabel="Acceleration")
        ax_u.legend()

        ax_DeltaJ = fig.add_subplot(2,3,4)
        ax_DeltaJ.grid(True, alpha=0.5)
        algo.plot_DeltaJ(ax_DeltaJ, summary_dict)
        ax_DeltaJ.axhline(tol_opt, color='k', linestyle='--', label='tol_opt')
        ax_DeltaJ.legend()

        ax_DeltaL = fig.add_subplot(2,3,5)
        ax_DeltaL.grid(True, alpha=0.5)
        algo.plot_chi(ax_DeltaL, summary_dict)
        ax_DeltaL.axhline(tol_feas, color='k', linestyle='--', label='tol_feas')
        ax_DeltaL.legend()

        ax = fig.add_subplot(2,3,6)
        for (_ts, _ys) in sols_ig:
            ax.plot(_ts, _ys[:,7], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ts, _ys[:,7], marker="o", ms=2, color='k')
        ax.grid(True, alpha=0.5)
        ax.set(xlabel="tau", ylabel="time")

        plt.tight_layout()
        fig.savefig(os.path.join(os.path.dirname(os.path.abspath(__file__)), "plots/scp_scipy_logmass_freetf_transfer.png"), dpi=300)
    return


if __name__ == "__main__":
    test_scp_scipy_logmass(get_plot=True)
    plt.show()


"""Integrator class"""

from collections.abc import Callable
import numpy as np
from scipy.integrate import solve_ivp

class ScipyIntegrator:
    """Integrator class using `scipy.integrate.solve_ivp`
    
    Args:
        nx (int): dimension of states
        nu (int): dimension of controls
        rhs (function): right-hand side function
        rhs_stm (function): right-hand side function for state transition matrix
        impulsive (bool): whether the dynamics are impulsive
        nv (int): dimensions corresponding to control norms to be augmented 
        method (str): integration method
        reltol (float): relative tolerance
        abstol (float): absolute tolerance
        args (tuple): additional arguments for the right-hand side function
    """
    def __init__(self, nx, nu, rhs: Callable, rhs_stm: Callable, impulsive=True, nv=0, method='RK45', reltol=1e-12, abstol=1e-12, args=None):
        """Initialize the integrator"""
        self.nx = nx
        self.nu = nu
        self.nv = nv
        self.rhs = rhs
        self.rhs_stm = rhs_stm
        self.impulsive = impulsive
        self.method = method
        self.reltol = reltol
        self.abstol = abstol
        self.args = args
        if impulsive is False and self.nu > 0:
            assert len(self.args[-1]) == self.nu + self.nv,\
                f"last argument must be a place-holder for control with length {self.nu + self.nv}, but got {len(self.args[-1])}; if only control is needed, pass as `args = ([0.0,],)`"
        return

    def solve(self, tspan, x0, u=None, stm=False, t_eval=None, args=None, get_ODESolution=False, dense_output=False):
        """Solve initial value problem

        Args:
            tspan (tuple): time span
            x0 (np.array): initial state
            stm (bool): whether to solve for state transition matrix
            t_eval (np.array): evaluation times
            args (tuple): additional arguments for the right-hand side function
            get_ODESolution (bool): whether to return an `ODESolution` object
            dense_output (bool): whether to return a dense output
        
        Returns:
            (tuple or ODESolution):
                if `get_ODESolution` is False, return a tuple of times and state with shape `N-by-nx`
                if `get_ODESolution` is True, return an `ODESolution` object
        """
        assert len(x0) == self.nx, f"x0 must be of length {self.nx}, but got {len(x0)}"

        if args is None:
            args = self.args
        if (u is not None) and (self.impulsive is False):
            args[-1][:] = u[:]

        if stm is False:
            sol = solve_ivp(self.rhs, tspan, x0,
                            t_eval=t_eval, method=self.method, rtol=self.reltol, atol=self.abstol, args=args, dense_output=dense_output)
        else:
            if self.impulsive is True:
                x0_stm = np.concatenate((x0, np.eye(self.nx).flatten()))
            else:
                x0_stm = np.concatenate((x0, np.eye(self.nx).flatten(), np.zeros(self.nx*(self.nu+self.nv))))
            sol = solve_ivp(self.rhs_stm, tspan, x0_stm, t_eval=t_eval, method=self.method, rtol=self.reltol, atol=self.abstol, args=args)
        if get_ODESolution is True:
            return sol
        else:
            return sol.t, sol.y.T
    


"""Test integrator class"""

import matplotlib.pyplot as plt
import numpy as np

import os
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))

import scocp

def test_scipy_integrator_cr3bp_impulsive():
    """Test `ScipyIntegrator` class for impulsive CR3BP dynamics"""
    mu = 1.215058560962404e-02
    integrator = scocp.ScipyIntegrator(nx=6, nu=3, rhs=scocp.rhs_cr3bp, rhs_stm=scocp.rhs_cr3bp_stm, args=(mu,),
                                       method='DOP853', reltol=1e-12, abstol=1e-12)

    # propagate uncontrolled and controlled dynamics
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0])
    period_0 = 2.3538670417546639E+00
    sol_lpo0 = integrator.solve([0.0, period_0], x0, stm=True, get_ODESolution=True)

    assert sol_lpo0.y[0:6,-1].shape == (6,)
    assert np.max(np.abs((sol_lpo0.y[0:6,-1] - x0))) < 1e-11
    return


def test_scipy_integrator_cr3bp_continuous():
    """Test `ScipyIntegrator` class for continuous CR3BP dynamics"""
    mu = 1.215058560962404e-02
    integrator = scocp.ScipyIntegrator(nx=6, nu=3, rhs=scocp.control_rhs_cr3bp, rhs_stm=scocp.control_rhs_cr3bp_stm,
                                       impulsive=False, args=(mu,[0.0,0.0,0.0]),
                                       method='DOP853', reltol=1e-12, abstol=1e-12)

    # propagate uncontrolled and controlled dynamics
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0])
    period_0 = 2.3538670417546639E+00
    sol_lpo0 = integrator.solve([0.0, period_0], x0, stm=False, get_ODESolution=True)

    assert sol_lpo0.y[0:6,-1].shape == (6,)
    assert np.max(np.abs((sol_lpo0.y[0:6,-1] - x0))) < 1e-11
    return


def test_scipy_integrator_twobody_mass_freetf():
    """Test `ScipyIntegrator` class for twobody dynamics with mass and time"""
    mu = 1.0
    c1 = 1e-2
    c2 = 1e-2
    integrator = scocp.ScipyIntegrator(nx=8, nu=5, rhs=scocp.control_rhs_twobody_mass_freetf,
                                       rhs_stm=scocp.control_rhs_twobody_mass_freetf_stm,
                                       impulsive=False, args=((mu, c1, c2), [0.0, 0.0, 0.0, 1.0, 0.0]),
                                       method='DOP853', reltol=1e-12, abstol=1e-12)
    
    
    x0 = np.array([1.0, 0.0, 0.0, 0.0, 1.0, 0.02, 1.0, 0.0])   # [r,v,mass,t]
    tspan = [0.0, 2*np.pi]
    u = np.array([0.04, 0.5, -0.2, tspan[1], np.sqrt(0.04**2 + 0.05**2 + 0.02**2)])
    sol_lpo0 = integrator.solve(tspan, x0, u=u, stm=True, get_ODESolution=True)

    fig = plt.figure(figsize=(10,5))
    ax = fig.add_subplot(1,2,1, projection='3d')
    ax.plot(sol_lpo0.y[0,:], sol_lpo0.y[1,:], sol_lpo0.y[2,:])
    ax = fig.add_subplot(1,2,2)
    ax.plot(sol_lpo0.y[7,:], sol_lpo0.y[6,:])
    plt.tight_layout()
    return


if __name__ == "__main__":
    test_scipy_integrator_twobody_mass_freetf()
    plt.show()



"""Impulsive rendezvous problems with fixed final time"""

from collections.abc import Callable
import cvxpy as cp
import numpy as np

from ._scocp_impulsive import ImpulsiveControlSCOCP
from .._misc import get_augmented_lagrangian_penalty


class FixedTimeImpulsiveRdv(ImpulsiveControlSCOCP):
    """Fixed-time impulsive rendezvous problem class"""
    def __init__(self, x0, xf, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x0 = x0
        self.xf = xf
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        return np.sum(vs)
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, vs, ys, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        
        xs = cp.Variable((N, nx), name='state')
        us = cp.Variable((N, nu), name='control')
        vs = cp.Variable((N, 1), name='Gamma')
        xis = cp.Variable((Nseg,nx), name='xi')         # slack for dynamics
        
        penalty = get_augmented_lagrangian_penalty(self.weight, xis, self.lmb_dynamics)
        objective_func = cp.sum(vs) + penalty
        constraints_objsoc = [cp.SOC(vs[i,0], us[i,:]) for i in range(N)]

        constraints_dyn = [
            xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,:] @ us[i,:] + self.Phi_c[i,:] + xis[i,:]
            for i in range(Nseg)
        ]

        constraints_trustregion = [
            xs[i,:] - xbar[i,:] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,:] - xbar[i,:] >= -self.trust_region_radius for i in range(N)
        ]

        constraints_initial = [xs[0,:] == self.x0]
        constraints_final   = [xs[-1,0:3] == self.xf[0:3], 
                               xs[-1,3:6] + us[-1,:] == self.xf[3:6]]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_objsoc + constraints_dyn + constraints_trustregion + constraints_initial + constraints_final)
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return xs.value, us.value, vs.value, None, xis.value, None, None
    



"""CR3BP equations of motion with scipy"""

from numba import njit, float64
import numpy as np


@njit(float64[:,:](float64[:], float64), cache=True)
def gravity_gradient_cr3bp(rvec, mu):
    """Compute gravity gradient matrix for CR3BP in the rotating frame.

    Args:
        rvec (np.array): position vector of spacecraft
        mu (float): CR3BP parameter, i.e. scaled mass of secondary body

    Returns:
        np.array: 3-by-3 gravity gradient matrix
    """
    # unpack rvec
    x,y,z = rvec
    r1 = np.sqrt((x+mu)**2 + y**2 + z**2)
    r2 = np.sqrt((x-(1-mu))**2 + y**2 + z**2)

    # define entries of the gravity gradient matrix
    r1_5 = r1**5
    r2_5 = r2**5
    G00 = 3*(1-mu)*(mu+x)**2/r1_5- (1-mu)/r1**3 + 3*mu*(x-1+mu)**2/r2_5 - mu/r2**3 + 1
    G01 = 3*(1-mu)*y*(mu+x)/r1_5+ 3*mu*y*(x-1+mu)/r2_5
    G02 = 3*(1-mu)*z*(mu+x)/r1_5+ 3*mu*z*(x-1+mu)/r2_5

    G10 = 3*(1-mu)*y*(mu+x)/r1_5+ 3*mu*y*(x-1+mu)/r2_5
    G11 = 3*(1-mu)*y**2/r1_5+ 3*mu*y**2/r2_5 - (1-mu)/r1**3 - mu/r2**3 + 1
    G12 = 3*(1-mu)*y*z/r1_5+ 3*mu*y*z/r2_5

    G20 = 3*(1-mu)*z*(mu+x)/r1_5+ 3*mu*z*(x-1+mu)/r2_5
    G21 = 3*(1-mu)*y*z/r1_5+ 3*mu*y*z/r2_5
    G22 = 3*(1-mu)*z**2/r1_5+ 3*mu*z**2/r2_5 - (1-mu)/r1**3 - mu/r2**3
    return np.array([[G00, G01, G02], [G10, G11, G12], [G20, G21, G22]])


@njit(float64[:](float64, float64[:], float64), cache=True)
def rhs_cr3bp(t, state, mu):
    """Equation of motion in CR3BP, formulated for scipy.integrate.solve=ivp(), compatible with njit

    Args:
        t (float): time
        state (np.array): 1D array of Cartesian state, length 6
        mu (float): CR3BP parameter

    Returns:
        (np.array): 1D array of derivative of Cartesian state
    """
    # unpack positions
    x = state[0]
    y = state[1]
    z = state[2]
    # unpack velocities
    vx = state[3]
    vy = state[4]
    vz = state[5]
    # compute radii to each primary
    r1 = np.sqrt((x + mu) ** 2 + y ** 2 + z ** 2)
    r2 = np.sqrt((x - 1 + mu) ** 2 + y ** 2 + z ** 2)
    # setup vector for dX/dt
    deriv = np.zeros((6,))
    # position derivatives
    deriv[0] = vx
    deriv[1] = vy
    deriv[2] = vz
    # velocity derivatives
    deriv[3] = (
        2 * vy + x - ((1 - mu) / r1 ** 3) * (mu + x) + (mu / r2 ** 3) * (1 - mu - x)
    )
    deriv[4] = -2 * vx + y - ((1 - mu) / r1 ** 3) * y - (mu / r2 ** 3) * y
    deriv[5] = -((1 - mu) / r1 ** 3) * z - (mu / r2 ** 3) * z
    return deriv


def rhs_cr3bp_stm(t, state, mu):
    """Equation of motion in CR3BP in the rotating frame with STM.
    This function is written for `scipy.integrate.solve=ivp()` and is compatible with njit.
    """
    # state derivative
    deriv = np.zeros(42)
    deriv[0:6] = rhs_cr3bp(t, state[0:6], mu)
    
    # derivative of STM
    A = np.zeros((6,6))
    A[0:3,3:6] = np.eye(3)
    A[3,4] = 2
    A[4,3] = -2
    A[3:6,0:3] = gravity_gradient_cr3bp(state[0:3], mu)
    deriv[6:42] = np.dot(A, state[6:].reshape(6,6)).reshape(36,)
    return deriv


def control_rhs_cr3bp(t, state, mu, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame"""
    # derivative of state
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv = rhs_cr3bp(t, state[0:6], mu) + B @ u[0:3]
    return deriv


def control_rhs_cr3bp_stm(t, state, mu, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame with STM"""
    # derivative of state
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv = np.zeros(60)    # 6 + 6*6 + 6*3
    deriv[0:6] = rhs_cr3bp(t, state[0:6], mu) + B @ u[0:3]
    
    # derivative of STM
    Phi_A = state[6:42].reshape(6,6)
    A = np.zeros((6,6))
    A[0:3,3:6] = np.eye(3)
    A[3,4] = 2
    A[4,3] = -2
    A[3:6,0:3] = gravity_gradient_cr3bp(state[0:3], mu)
    deriv[6:42] = np.dot(A, Phi_A).reshape(36,)

    # derivative of control sensitivity
    Phi_B = state[42:60].reshape(6,3)
    deriv[42:60] = (np.dot(A, Phi_B) + B).reshape(18,)
    return deriv


def control_rhs_cr3bp_logmass(t, state, parameters, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame"""
    # unpack parameters
    mu, cex = parameters
    # derivative of state
    B = np.array([
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0, -1/cex],
    ])
    deriv = np.zeros(7)
    deriv[0:6] = rhs_cr3bp(t, state[0:6], mu)
    deriv += B @ u
    return deriv


def control_rhs_cr3bp_logmass_stm(t, state, parameters, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame with STM"""
    # derivative of state
    mu, cex = parameters
    B = np.array([
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0, 0.0],
        [0.0, 1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0, 0.0],
        [0.0, 0.0, 0.0, -1/cex],
    ])
    deriv = np.zeros(84)    # 7 + 7*7 + 7*4
    deriv[0:6] = rhs_cr3bp(t, state[0:6], mu)
    deriv[0:7] += B @ u[:]
    
    # derivative of STM
    Phi_A = state[7:56].reshape(7,7)
    A = np.zeros((7,7))
    A[0:3,3:6] = np.eye(3)
    A[3,4] = 2
    A[4,3] = -2
    A[3:6,0:3] = gravity_gradient_cr3bp(state[0:3], mu)
    deriv[7:56] = np.dot(A, Phi_A).reshape(49,)

    # derivative of control sensitivity
    Phi_B = state[56:84].reshape(7,4)
    deriv[56:84] = (np.dot(A, Phi_B) + B).reshape(28,)
    return deriv


def control_rhs_cr3bp_freetf(tau, state, mu, u):
    """Equation of motion in CR3BP with free final time
    state = [x,y,z,vx,vy,vz,t]
    u = [ax,ay,az,s] where s is the dilation factor
    """
    t = state[6]
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv = np.zeros(7,)
    deriv[0:6] = u[3] * (rhs_cr3bp(t, state[0:6], mu) + B @ u[0:3])
    deriv[6]   = u[3]
    return deriv


def control_rhs_cr3bp_freetf_stm(tau, state, mu, u):
    """Equation of motion in CR3BP with free final time
    state = [x,y,z,vx,vy,vz,t]
    u = [ax,ay,az,s] where s is the dilation factor
    """
    # derivative of state
    deriv = np.zeros(84)    # 7 + 7*7 + 7*4
    deriv[0:7] = control_rhs_cr3bp_freetf(tau, state[0:7], mu, u)
    
    # derivative of STM
    Phi_A = state[7:56].reshape(7,7)
    A = np.zeros((7,7))
    A[0:3,3:6] = np.eye(3)
    A[3,4] = 2
    A[4,3] = -2
    A[3:6,0:3] = gravity_gradient_cr3bp(state[0:3], mu)
    deriv[7:56] = np.dot(u[3] * A, Phi_A).reshape(49,)          # note: multiply A by time-dilation factor

    # derivative of control sensitivity
    f_eval = control_rhs_cr3bp_freetf(tau, state[0:7], mu, u)[0:6]/u[3]
    B_accel = np.concatenate((np.concatenate((np.zeros((3,3)), u[3]*np.eye(3))), f_eval.reshape(-1,1)), axis=1)
    B = np.concatenate((B_accel, np.array([0.0, 0.0, 0.0, 1.0]).reshape(1,-1)))
    Phi_B = state[56:84].reshape(7,4)
    deriv[56:84] = (np.dot(u[3] * A, Phi_B) + B).reshape(28,)
    return deriv


def control_rhs_cr3bp_logmass_freetf(tau, state, parameters, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame
    state = [x,y,z,vx,vy,vz,log(mass),t]
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    """
    # unpack parameters
    mu, cex = parameters
    # derivative of state
    deriv = np.zeros(8,)
    B = np.concatenate((np.zeros((3,3)), np.eye(3)))
    deriv[0:6] = u[3] * (rhs_cr3bp(state[6], state[0:6], mu) + B @ u[0:3])
    deriv[6]   = u[3] * u[4] * (-1/cex)     # control on log(mass), multiplied by time-dilation factor
    deriv[7]   = u[3]                       # control on t(tau)
    return deriv


def control_rhs_cr3bp_logmass_freetf_stm(tau, state, parameters, u):
    """Equation of motion in CR3BP with continuous control in the rotating frame with STM
    state = [x,y,z,vx,vy,vz,log(mass),t] + STM.flatten()
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    """
    # derivative of state
    mu, cex = parameters
    deriv = np.zeros(112)    # 8 + 8*8 + 8*5
    deriv[0:8] = control_rhs_cr3bp_logmass_freetf(tau, state[0:8], parameters, u)
    
    # derivative of STM
    Phi_A = state[8:72].reshape(8,8)
    A = np.zeros((8,8))
    A[0:3,3:6] = np.eye(3)
    A[3,4] = 2
    A[4,3] = -2
    A[3:6,0:3] = gravity_gradient_cr3bp(state[0:3], mu)
    deriv[8:72] = np.dot(u[3] * A, Phi_A).reshape(64,)

    # derivative of control sensitivity
    B = np.zeros((8,5))
    B[3:6,0:3] = u[3] * np.eye(3)
    B[0:8,3] = control_rhs_cr3bp_logmass_freetf(tau, state[0:8], parameters, u)/u[3]
    B[6,4] = -u[3]/cex
    Phi_B = state[72:112].reshape(8,5)
    deriv[72:112] = (np.dot(u[3] * A, Phi_B) + B).reshape(40,)
    return deriv


"""SOCP with continuous control"""

from ._scocp_continuous import (
    ContinuousControlSCOCP,
)
from ._rdv_fixed import (
    FixedTimeContinuousRdv,
    FixedTimeContinuousRdvLogMass,
)
from ._rdv_free import (
    FreeTimeContinuousRdv,
    FreeTimeContinuousRdvLogMass,
)
from ._rdv_free_moving import (
    FreeTimeContinuousMovingTargetRdvLogMass,
    FreeTimeContinuousMovingTargetRdvMass,
)
from ._ballistic import (
    FixedTimeBallisticTrajectory,
)



"""Using scocp for multiple shooting, without any control"""

import cvxpy as cp
import numpy as np
import matplotlib.pyplot as plt

import os
import sys
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))

import scocp

def test_scp_scipy_nocontrol(get_plot = False):
    """Test SCP without control"""
    mu = 1.215058560962404e-02
    integrator = scocp.ScipyIntegrator(nx=6, nu=0, rhs=scocp.rhs_cr3bp, rhs_stm=scocp.rhs_cr3bp_stm, args=(mu,),
                                       impulsive=True, method='DOP853', reltol=1e-12, abstol=1e-12)
    
    # initial state of true, sought orbit
    x0 = np.array([
        1.0809931218390707E+00,
        0.0,
        -2.0235953267405354E-01,
        0.0,
        -1.9895001215078018E-01,
        0.0])
    period = 2.3538670417546639E+00

    # construct discretized times and evaluate true solution along it for testing
    N = 10
    times = np.linspace(0, period, N)
    sol = integrator.solve([0, period], x0, get_ODESolution=True, t_eval=times)

    # create initial guess by perturbing true solution
    xbar = np.array([
        x_true + 0.05 * np.random.randn(6)
        for x_true in sol.y[:6, :].T
    ])
    xbar[0] = x0
    xbar[-1] = x0
    print(f"xbar.shape = {xbar.shape}")

    # create problem for a periodic trajectory
    problem = scocp.FixedTimeBallisticTrajectory(x0, x0, 0.0, integrator, times)

    # solve subproblem
    ubar = np.zeros((N-1,0))
    vbar = np.zeros((N-1,0))
    problem.solve_convex_problem(xbar, ubar, vbar)
    assert problem.cp_status == "optimal", f"problem.cp_status = {problem.cp_status}"

    # setup algorithm & solve
    tol_feas = 1e-10
    tol_opt = 1e-4
    algo = scocp.SCvxStar(problem, tol_opt=tol_opt, tol_feas=tol_feas)
    solution = algo.solve(
        xbar,
        ubar,
        vbar,
        maxiter = 100,
        verbose = True
    )
    xopt, uopt, vopt, yopt, sols, summary_dict = solution.x, solution.u, solution.v, solution.y, solution.sols, solution.summary_dict
    # assert summary_dict["status"] == "Optimal"
    # assert summary_dict["chi"][-1] <= tol_feas
    print(f"ubar.shape = {ubar.shape}, uopt.shape = {uopt.shape}")
    
    # evaluate solution
    if (get_plot is True) and (summary_dict["status"] != "CPFailed"):
        _, sols_ig = problem.evaluate_nonlinear_dynamics(xbar, ubar, vbar, steps=5)
    
        # plot results
        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(2,2,1,projection='3d')
        for (_ts, _ys) in sols_ig:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], '--', color='grey')
        for (_ts, _ys) in sols:
            ax.plot(_ys[:,0], _ys[:,1], _ys[:,2], 'b-')

        ax.scatter(x0[0], x0[1], x0[2], marker='x', color='k', label='Initial state')
        # ax.scatter(xf[0], xf[1], xf[2], marker='o', color='k', label='Final state')
        ax.set_aspect('equal')
        ax.legend()

        # ax_u = fig.add_subplot(2,2,2)
        # ax_u.grid(True, alpha=0.5)
        # ax_u.step(times, np.concatenate((vopt[:,0], [0.0])), label="Gamma", where='post', color='k')
        # ax_u.set(xlabel="Time", ylabel="Control")
        # ax_u.legend()

        ax_DeltaJ = fig.add_subplot(2,2,3)
        ax_DeltaJ.grid(True, alpha=0.5)
        algo.plot_DeltaJ(ax_DeltaJ, summary_dict)
        ax_DeltaJ.axhline(tol_opt, color='k', linestyle='--', label='tol_opt')
        ax_DeltaJ.legend()

        ax_DeltaL = fig.add_subplot(2,2,4)
        ax_DeltaL.grid(True, alpha=0.5)
        algo.plot_chi(ax_DeltaL, summary_dict)
        ax_DeltaL.axhline(tol_feas, color='k', linestyle='--', label='tol_feas')
        ax_DeltaL.legend()

        plt.tight_layout()
    return


if __name__ == "__main__":
    test_scp_scipy_nocontrol(get_plot = True)
    plt.show()


"""Sequentially convexified optimal control problem (SCOCP) for continuous dynamics"""

import cvxpy as cp
import numpy as np

from .._misc import get_augmented_lagrangian_penalty


class ContinuousControlSCOCP:
    """Sequentially convexified optimal control problem (SCOCP) for continuous dynamics
    
    Args:
        integrator (obj): integrator object
        times (np.array): time grid
        ng (int): number of nonlinear equality constraints, excluding dynamics constraints
        nh (int): number of nonlinear inequality constraints
        ny (int): number of other variables
        augment_Gamma (bool): whether to augment the control with the constraint vector when integrating the dynamics
        weight (float): weight of the objective function
        trust_region_radius (float): trust region radius
        solver (str): solver to use
        verbose_solver (bool): whether to print verbose output
    """
    def __init__(
        self,
        integrator,
        times,
        ng: int = 0,
        nh: int = 0,
        ny: int = 0,
        augment_Gamma: bool = False,
        weight: float = 1e2,
        trust_region_radius: float = 0.1,
        solver = cp.CLARABEL,
        verbose_solver: bool = False,
        impulsive_B = None,
    ):
        #assert integrator.impulsive is False, "Continuous control problem must be initialized with an integrator for continuous dynamics"
        assert weight >= 0.0, f"weight must be non-negative, but given {weight}"
        self.integrator = integrator
        self.times = times
        self.N = len(times)
        self.ng_dyn = self.integrator.nx * (self.N - 1)
        self.ng = ng
        self.nh = nh
        self.ny = ny
        self.weight_initial = weight
        self.trust_region_radius_initial = trust_region_radius
        self.solver = solver
        self.verbose_solver = verbose_solver
        self.augment_Gamma = augment_Gamma

        if impulsive_B is None:
            self.impulsive_B = np.zeros((self.integrator.nx, self.integrator.nu))
        else:
            self.impulsive_B = impulsive_B
        
        # reset the problem
        self.reset()
        return
    
    def reset(self):
        """Reset problem parameters and storages"""
        self.weight = self.weight_initial
        self.trust_region_radius = self.trust_region_radius_initial
        self.cp_status = "not_solved"
        Nseg = self.N - 1
        if self.augment_Gamma:
            self.Phi_A = np.zeros((Nseg,self.integrator.nx,self.integrator.nx))
            self.Phi_B = np.zeros((Nseg,self.integrator.nx,self.integrator.nu+1))
            self.Phi_c = np.zeros((Nseg,self.integrator.nx))
            assert self.integrator.nv >= 1,\
                "When formulating SCOCP with `augment_Gamma=True`, `integrator.nv` must be set to at least 1"
        else:
            self.Phi_A = np.zeros((Nseg,self.integrator.nx,self.integrator.nx))
            self.Phi_B = np.zeros((Nseg,self.integrator.nx,self.integrator.nu))
            self.Phi_c = np.zeros((Nseg,self.integrator.nx))

        # initialize multipliers
        self.lmb_dynamics = np.zeros((Nseg,self.integrator.nx))
        self.lmb_eq       = np.zeros(self.ng)
        self.lmb_ineq     = np.zeros(self.nh)
        return
    
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        raise NotImplementedError("Subproblem must be implemented by inherited class!")
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar = None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
            ybar (np.array): `(self.ny,)` other reference variables
        
        Returns:
            (tuple): np.array values of xs, us, gs, xi_dyn, xi_eq, zeta_ineq
        """
        raise NotImplementedError("Subproblem must be implemented by inherited class!")
    
    def build_linear_model(self, xbar, ubar, vbar):
        """Construct linear model for dynamics multiple-shooting constraints within SCP algorithm
        This function computes the Phi_A, Phi_B, and Phi_c matrices and stores them in the class attributes.

        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of constraint history
        """
        assert xbar.shape == (self.N, self.integrator.nx),\
            f"Given incorrect xbar shape {xbar.shape}; should be {(self.N, self.integrator.nx)}"
        assert ubar.shape == (self.N-1, self.integrator.nu),\
            f"Given incorrect ubar shape {ubar.shape}; should be {(self.N-1, self.integrator.nu)}"
        if vbar is not None:
            if self.integrator.nv == 0 and self.integrator.nu > 0:
                assert vbar.shape == (self.N-1,1), f"Given incorrect vbar shape {vbar.shape}; should be {(self.N-1,1)}"
            else:
                assert vbar.shape == (self.N-1, self.integrator.nv),\
                    f"Given incorrect vbar shape {vbar.shape}; should be {(self.N-1, self.integrator.nv)}"
        i_PhiA_end = self.integrator.nx + self.integrator.nx * self.integrator.nx
        for i,ti in enumerate(self.times[:-1]):
            _tspan = (ti, self.times[i+1])
            _x0 = xbar[i,:] + self.impulsive_B @ ubar[i,:]

            if self.augment_Gamma:
                _ubar = np.concatenate((ubar[i,:], vbar[i,:]))
            else:
                _ubar = ubar[i,:]
            _, _ys = self.integrator.solve(_tspan, _x0, u=_ubar, stm=True)

            xf  = _ys[-1,0:self.integrator.nx]
            self.Phi_A[i,:,:] = _ys[-1,self.integrator.nx:i_PhiA_end].reshape(self.integrator.nx,self.integrator.nx)
            if self.augment_Gamma:
                self.Phi_B[i,:,:] = _ys[-1,i_PhiA_end:].reshape(self.integrator.nx,self.integrator.nu+1)
            else:
                self.Phi_B[i,:,:] = _ys[-1,i_PhiA_end:].reshape(self.integrator.nx,self.integrator.nu)
            self.Phi_c[i,:]   = xf - self.Phi_A[i,:,:] @ (xbar[i,:] + self.impulsive_B @ ubar[i,:]) - self.Phi_B[i,:,:] @ _ubar
        return
        
    def evaluate_nonlinear_dynamics(self, xs, us, vs, stm = False, steps = None):
        """Evaluate nonlinear dynamics along given state and control history
        
        Args:
            integrator (obj): integrator object
            times (np.array): time grid
            vs (np.array): state history
            ubar (np.array): control history
            stm (bool): whether to propagate STMs, defaults to False
        """
        assert xs.shape == (self.N,self.integrator.nx)
        assert us.shape == (self.N-1,self.integrator.nu)

        sols = []
        geq_nl = np.zeros((self.N-1,self.integrator.nx))
        for i,ti in enumerate(self.times[:-1]):
            _tspan = (ti, self.times[i+1])
            _x0 = xs[i,:] + self.impulsive_B @ us[i,:]
            if steps is None:
                t_eval = None
            else:
                t_eval = np.linspace(ti, self.times[i+1], steps)
            if self.augment_Gamma:
                _ts, _ys = self.integrator.solve(_tspan, _x0, u=np.concatenate((us[i,:], vs[i,:])), stm=stm, t_eval=t_eval)
            else:
                _ts, _ys = self.integrator.solve(_tspan, _x0, u=us[i,:], stm=stm, t_eval=t_eval)
            sols.append([_ts,_ys])
            geq_nl[i,:] = xs[i+1,:] - _ys[-1,0:self.integrator.nx]
        return geq_nl, sols
    
    def evaluate_nonlinear_constraints(self, xs, us, gs, ys=None):
        """Evaluate nonlinear constraints
        
        Returns:
            (tuple): tuple of 1D arrays of nonlinear equality and inequality constraints
        """
        return np.zeros(self.ng), np.zeros(self.nh)



"""Free-time problems"""

from collections.abc import Callable
import cvxpy as cp
import numpy as np

from ._scocp_continuous import ContinuousControlSCOCP
from .._misc import get_augmented_lagrangian_penalty, MovingTarget



class FreeTimeContinuousRdv(ContinuousControlSCOCP):
    """Free-time continuous rendezvous problem class
    A good initial guess for the dilation factor `s` is the guessed TOF, since `dt/dtau = s`.
    
    Args:
        x0 (np.array): initial state
        xf (np.array): final state
        umax (float): maximum control magnitude
        tf_bounds (list): bounds on the final time, given as [tf_min, tf_max]
        s_bounds (list): bounds on time dilation factor, given as [s_min, s_max]
    """
    def __init__(self, x0, xf, umax, tf_bounds, s_bounds, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert abs(self.times[0]  - 0.0) < 1e-14, f"self.times[0] must be 0.0, but given {self.times[0]}"
        assert abs(self.times[-1] - 1.0) < 1e-14, f"self.times[-1] must be 1.0, but given {self.times[-1]}"
        assert s_bounds[0] > 0.0, f"s_bounds[0] must be greater than 0.0, but given {s_bounds[0]}"
        assert s_bounds[0] < s_bounds[1], f"s_bounds[0] must be less than s_bounds[1], but given {s_bounds[0]} and {s_bounds[1]}"
        self.x0 = x0
        self.xf = xf
        self.umax = umax
        self.tf_bounds = tf_bounds
        self.s_bounds = s_bounds
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        dts = np.diff(self.times)
        return np.sum(vs.T @ dts)
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, gs, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        assert nx == self.integrator.nx, f"xbar.shape[1] = {xbar.shape[1]} must match integrator.nx = {self.integrator.nx}"
        assert nu == self.integrator.nu, f"ubar.shape[1] = {ubar.shape[1]} must match integrator.nu = {self.integrator.nu}"
        
        xs = cp.Variable((N, nx), name='state')
        us = cp.Variable((Nseg, nu), name='control')
        vs = cp.Variable((Nseg, 1), name='Gamma')
        xis_dyn = cp.Variable((Nseg,nx), name='xi_dyn')         # slack for dynamics
        
        penalty = get_augmented_lagrangian_penalty(self.weight, xis_dyn, self.lmb_dynamics)
        dts = np.diff(self.times)
        objective_func = cp.sum(vs.T @ dts) + penalty
        constraints_objsoc = [cp.SOC(vs[i,0], us[i,0:3]) for i in range(N-1)]

        if self.augment_Gamma:
            constraints_dyn = [
                xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,:] @ np.concatenate([us[i,:], vs[i,:]]) + self.Phi_c[i,:] + xis_dyn[i,:]
                for i in range(Nseg)
            ]
        else:
            constraints_dyn = [
                xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,:] @ us[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
                for i in range(Nseg)
            ]

        constraints_trustregion = [
            xs[i,0:6] - xbar[i,0:6] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,0:6] - xbar[i,0:6] >= -self.trust_region_radius for i in range(N)
        ]

        constraints_initial = [xs[0,0:6]  == self.x0[0:6]]
        constraints_final   = [xs[-1,0:3] == self.xf[0:3], 
                               xs[-1,3:6] == self.xf[3:6]]
        
        constraint_t0 = [xs[0,6] == 0.0]
        constraints_tf      = [self.tf_bounds[0] <= xs[-1,6],
                               xs[-1,6] <= self.tf_bounds[1]]
        constraints_s       = [self.s_bounds[0] <= us[i,3] for i in range(Nseg)] + [us[i,3] <= self.s_bounds[1] for i in range(Nseg)]

        constraints_control = [
            vs[i,0] <= self.umax for i in range(Nseg)
        ]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_objsoc + constraints_dyn + constraints_trustregion +\
            constraints_initial + constraints_final + constraints_control +\
            constraint_t0 + constraints_tf + constraints_s
        )
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return xs.value, us.value, vs.value, None, xis_dyn.value, None, None


class FreeTimeContinuousRdvLogMass(ContinuousControlSCOCP):
    """Free-time continuous rendezvous problem class with log-mass dynamics
    
    Note the ordering expected for the state and the control vectors: 

    state = [x,y,z,vx,vy,vz,log(mass),t]
    u = [ax, ay, az, s, Gamma] where s is the dilation factor, Gamma is the control magnitude (at convergence)
    
    """
    def __init__(self, x0, xf, Tmax, tf_bounds, s_bounds, N, *args, **kwargs):
        super().__init__(nh = N - 1, *args, **kwargs)
        assert len(x0) == 7
        assert len(xf) >= 6
        assert abs(self.times[0]  - 0.0) < 1e-14, f"self.times[0] must be 0.0, but given {self.times[0]}"
        assert abs(self.times[-1] - 1.0) < 1e-14, f"self.times[-1] must be 1.0, but given {self.times[-1]}"
        assert s_bounds[0] > 0.0, f"s_bounds[0] must be greater than 0.0, but given {s_bounds[0]}"
        assert s_bounds[0] < s_bounds[1], f"s_bounds[0] must be less than s_bounds[1], but given {s_bounds[0]} and {s_bounds[1]}"
        self.x0 = x0
        self.xf = xf
        self.Tmax = Tmax
        self.tf_bounds = tf_bounds
        self.s_bounds = s_bounds
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        return -xs[-1,6]
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, gs, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        
        xs = cp.Variable((N, nx), name='state')
        us = cp.Variable((Nseg, nu), name='control')
        vs = cp.Variable((Nseg, 1), name='Gamma')
        xis_dyn = cp.Variable((Nseg,nx), name='xi_dyn')         # slack for dynamics
        zetas = cp.Variable((Nseg,), name='zeta')     # slack for non-convex inequality
        
        penalty = get_augmented_lagrangian_penalty(self.weight, xis_dyn, self.lmb_dynamics, zeta=zetas, lmb_ineq=self.lmb_ineq)
        objective_func = -xs[-1,6] + penalty
        constraints_objsoc = [cp.SOC(vs[i,0], us[i,0:3]) for i in range(N-1)]
        
        constraints_dyn = [
            xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,0:4] @ us[i,:] + self.Phi_B[i,:,4] * vs[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
            for i in range(Nseg)
        ]

        constraints_trustregion = [
            xs[i,:] - xbar[i,:] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,:] - xbar[i,:] >= -self.trust_region_radius for i in range(N)
        ]

        constraints_initial = [xs[0,0:7] == self.x0[0:7]]
        constraints_final   = [xs[-1,0:3] == self.xf[0:3], 
                               xs[-1,3:6] == self.xf[3:6]]
        
        constraint_t0 = [xs[0,7] == 0.0]
        constraints_tf      = [self.tf_bounds[0] <= xs[-1,7],
                               xs[-1,7] <= self.tf_bounds[1]]
        constraints_s       = [self.s_bounds[0] <= us[i,3] for i in range(Nseg)] + [us[i,3] <= self.s_bounds[1] for i in range(Nseg)]

        
        constraints_control = [
            vs[i,0] - self.Tmax * np.exp(-xbar[i,6]) * (1 - (xs[i,6] - xbar[i,6])) <= zetas[i]
            for i in range(Nseg)
        ]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_objsoc + constraints_dyn + constraints_trustregion +\
            constraints_initial + constraints_final + constraints_control +\
            constraint_t0 + constraints_tf + constraints_s)
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return xs.value, us.value, vs.value, None, xis_dyn.value, None, zetas.value
    
    def evaluate_nonlinear_constraints(self, xs, us, gs, ys=None):
        """Evaluate nonlinear constraints
        
        Returns:
            (tuple): tuple of 1D arrays of nonlinear equality and inequality constraints
        """
        h_ineq = np.array([
            max(gs[i,0] - self.Tmax * np.exp(-xs[i,6]), 0.0) for i in range(self.N-1)
        ])
        return np.zeros(self.ng), h_ineq



"""Ballistic trajectory design problem with fixed boundary conditions"""

from collections.abc import Callable
import cvxpy as cp
import numpy as np

from ._scocp_continuous import ContinuousControlSCOCP
from .._misc import get_augmented_lagrangian_penalty


class FixedTimeBallisticTrajectory(ContinuousControlSCOCP):
    """Fixed-time ballistic trajectory design problem class
    
    Initialize ubar and zbar with zeros of size (N-1,0)
    """
    def __init__(self, x0, xf, umax, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert len(x0) == self.integrator.nx
        assert len(xf) == self.integrator.nx
        assert self.integrator.nu == 0,\
            "Ballistic trajectory design problem must be initialized with an integrator for continuous dynamics"
        self.x0 = x0
        self.xf = xf
        self.umax = umax
        return
        
    def evaluate_objective(self, xs, us, vs, ys=None):
        """Evaluate the objective function"""
        return 1.0
    
    def solve_convex_problem(self, xbar, ubar, vbar, ybar=None):
        """Solve the convex subproblem
        
        Args:
            xbar (np.array): `(N, self.integrator.nx)` array of reference state history
            ubar (np.array): `(N-1, self.integrator.nu)` array of reference control history
            vbar (np.array): `(N-1, self.integrator.nv)` array of reference constraint history
        
        Returns:
            (tuple): np.array values of xs, us, vs, xi_dyn, xi_eq, zeta_ineq
        """
        N,nx = xbar.shape
        _,nu = ubar.shape
        Nseg = N - 1
        
        xs = cp.Variable((N, nx), name='state')
        xis_dyn = cp.Variable((Nseg,nx), name='xi_dyn')         # slack for dynamics
        
        penalty = get_augmented_lagrangian_penalty(self.weight, xis_dyn, self.lmb_dynamics)
        dts = np.diff(self.times)
        objective_func = 1.0 + penalty
        #constraints_objsoc = [cp.SOC(vs[i,0], us[i,:]) for i in range(N-1)]

        if self.augment_Gamma:
            constraints_dyn = [
                #xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_B[i,:,:] @ np.concatenate([us[i,:], vs[i,:]]) + self.Phi_c[i,:] + xis_dyn[i,:]
                xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
                for i in range(Nseg)
            ]
        else:
            constraints_dyn = [
                xs[i+1,:] == self.Phi_A[i,:,:] @ xs[i,:] + self.Phi_c[i,:] + xis_dyn[i,:]
                for i in range(Nseg)
            ]

        constraints_trustregion = [
            xs[i,:] - xbar[i,:] <= self.trust_region_radius for i in range(N)
        ] + [
            xs[i,:] - xbar[i,:] >= -self.trust_region_radius for i in range(N)
        ]

        constraints_initial = [xs[0,:] == self.x0]
        constraints_final   = [xs[-1,0:3] == self.xf[0:3], 
                               xs[-1,3:6] == self.xf[3:6]]

        convex_problem = cp.Problem(
            cp.Minimize(objective_func),
            constraints_dyn + constraints_trustregion + constraints_initial + constraints_final)
        convex_problem.solve(solver = self.solver, verbose = self.verbose_solver)
        self.cp_status = convex_problem.status
        return (
            xs.value,
            np.zeros((self.N-1,self.integrator.nu)),
            np.zeros((self.N-1,self.integrator.nu)),
            None, xis_dyn.value, None, None
        )
    
